"""Expression Atlas."""
import logging
import os
import sys
from collections import OrderedDict
from typing import List, Optional, Tuple

import pandas as pd
import xmltodict
from pandas.core.frame import DataFrame
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel.constants import DATA_DIR
from ebel.manager.orientdb import odb_meta, urls
from ebel.manager.rdbms.models import expression_atlas
from ebel.tools import get_standard_name

logger = logging.getLogger(__name__)


class ExpressionAtlas(odb_meta.Graph):
    """ExpressionAtlas."""

    def __init__(self, client: OrientDB = None):
        """Init ExpressionAtlas."""
        self.client = client
        self.biodb_name = "expression_atlas"
        self.urls = {"latest_data": urls.EXPRESSION_ATLAS_EXPERIMENTS}
        self.data_dir = os.path.join(DATA_DIR, self.biodb_name)

        super().__init__(
            tables_base=expression_atlas.Base,
            urls=self.urls,
            biodb_name=self.biodb_name,
        )

    def __len__(self):
        return self.number_of_generics

    def __contains__(self, item):
        # TODO: To be implemented
        return True

    def update(self):
        """Update ExpressionAtlas."""
        logger.info("Update ExpressionAtlas")
        downloaded = self.download()
        if downloaded["latest_data"]:
            self.extract_files()
            self.insert()

    def extract_files(self):
        """Extract relevant files."""
        os.chdir(self.data_dir)
        cmd_temp = "tar -xzf atlas-latest-data.tar.gz --wildcards --no-anchored '{}'"
        patterns = [
            "*.sdrf.txt",
            "*.condensed-sdrf.tsv",
            "*analytics.tsv",
            "*-configuration.xml",
            "*.idf.txt",
            "*.go.gsea.tsv",
            "*.interpro.gsea.tsv",
            "*.reactome.gsea.tsv",
        ]
        with tqdm(patterns) as t_patterns:
            for pattern in t_patterns:
                t_patterns.set_description(f"Extract files with pattern {pattern}")
                command = cmd_temp.format(pattern)
                os.system(command)

    def insert_data(self):
        """Class method."""
        pass

    def update_interactions(self) -> int:
        """Class method."""
        pass

    def insert_experiment(self, experiment_name: str, title: str) -> int:
        """Insert individual experiment into SQL database.

        Parameters
        ----------
        experiment_name : str
            Name of the Expression Atlas experiment.
        title : str
            Title of experiment.

        Returns
        -------
        Table ID of newly inserted experiment.
        """
        experiment = expression_atlas.Experiment(name=experiment_name, title=title)
        self.session.add(experiment)
        self.session.flush()
        self.session.commit()
        return experiment.id

    def insert(self):
        """Override insert method for SQL data insertion."""
        self.recreate_tables()
        data_folder = os.scandir(self.data_dir)
        for experiment_name in tqdm([(f.name) for f in data_folder if f.is_dir()]):
            try:
                df_configuration = self.get_configuration(experiment_name)
                if isinstance(df_configuration, pd.DataFrame):
                    df_idf = self.get_idf(experiment_name)
                    title = df_idf[df_idf.key_name == "investigation_title"].value.values[0]

                    experiment_id = self.insert_experiment(experiment_name, title)

                    groups_strs: Tuple[str, ...] = self.__insert_configuration(df_configuration, experiment_id)

                    self.__insert_idf(df_idf, experiment_id)
                    self.__insert_sdrf_condensed(experiment_id, experiment_name)
                    self.__insert_foldchange(experiment_id, experiment_name, groups_strs)
                    self.insert_gseas(experiment_id, experiment_name, groups_strs)

            except Exception as e:
                print(experiment_name)
                print(e)
                sys.exit()

    def __insert_foldchange(self, experiment_id: int, experiment_name: str, groups_strs: Tuple[str, ...]):
        df_log2foldchange = self.get_log2foldchange(experiment_name, groups_strs).set_index("group_comparison")
        df_group_comparison = self.get_df_group_comparison(experiment_id, groups_strs).set_index("group_comparison")
        df_log2foldchange.join(df_group_comparison).to_sql(
            expression_atlas.FoldChange.__tablename__,
            self.engine,
            if_exists="append",
            index=False,
        )

    def get_df_group_comparison(self, experiment_id: int, groups_strs: Tuple[str, ...]) -> pd.DataFrame:
        """Get group comparison IDs and group comparison columns for pairs of group strings.

        Parameters
        ----------
        experiment_id : int
            Experiment numerical ID.
        groups_strs : tuple
            Pairs of gene symbols.

        Returns
        -------
        Pandas DataFrame of 'group_comparison_id' and 'group_comparison'.
        """
        data = []
        for groups_str in groups_strs:
            group_comparison_id = (
                self.session.query(expression_atlas.GroupComparison.id)
                .filter_by(experiment_id=experiment_id, group_comparison=groups_str)
                .first()
                .id
            )
            data.append((group_comparison_id, groups_str))
        return pd.DataFrame(data, columns=["group_comparison_id", "group_comparison"])

    def __insert_configuration(self, df_configuration, experiment_id: int) -> Tuple[str, ...]:
        df_configuration["experiment_id"] = experiment_id
        df_configuration.to_sql(
            expression_atlas.GroupComparison.__tablename__,
            self.engine,
            if_exists="append",
            index=False,
        )
        groups_strs = tuple(df_configuration.group_comparison.values)
        self.session.flush()
        self.session.commit()
        return groups_strs

    def __insert_idf(self, df_idf: DataFrame, experiment_id: int):
        df_idf["experiment_id"] = experiment_id
        df_idf.to_sql(
            expression_atlas.Idf.__tablename__,
            self.engine,
            if_exists="append",
            index=False,
        )

    def __insert_sdrf_condensed(self, experiment_id: int, experiment_name: str):
        df = self.get_sdrf_condensed(experiment_name)
        # organisms = tuple(df[df.parameter == 'organism'].value.unique())
        df["experiment_id"] = experiment_id
        df.drop(columns=["experiment"], inplace=True)
        df.to_sql(
            expression_atlas.SdrfCondensed.__tablename__,
            self.engine,
            if_exists="append",
            index=False,
        )

    def insert_gseas(self, experiment_id: int, experiment_name: str, groups_strs: Tuple[str, ...]):
        """Insert Gene set enrichment analysis.

        For more information about parameters see https://www.bioconductor.org/packages/release/bioc/html/piano.html

        Args:
            experiment_id (int): [description]
            experiment_name (str): [description]
            groups_strs (Tuple[str, ...]): [description]
        """
        df = self.get_gseas(experiment_name, experiment_id, groups_strs)
        if isinstance(df, pd.DataFrame):
            df.to_sql(
                expression_atlas.Gsea.__tablename__,
                self.engine,
                if_exists="append",
                index=False,
            )

    def get_gseas(self, experiment_name: str, experiment_id: int, groups_strs: Tuple[str]) -> Optional[pd.DataFrame]:
        """Get GSEA data.

        Parameters
        ----------
        experiment_name : str
            Name of the Expression Atlas experiment.
        experiment_id : int
            Experiment numerical ID.
        groups_strs : tuple
            Pairs of gene symbols.

        Returns
        -------
        If GSEA informaiton is found, returns a pandas DataFrame detailing the GSEAs associated with an experiment.
        """
        dfs = []
        for groups_str in groups_strs:
            for gsea_type in ["go", "reactome", "interpro"]:
                df = self.get_gsea(experiment_name, groups_str, gsea_type)
                if isinstance(df, pd.DataFrame):
                    df["gsea_type"] = gsea_type
                    df["group_comparison_id"] = self.__get_group_comparison_id(groups_str, experiment_id)
                    dfs.append(df[df.p_adj_non_dir <= 0.05])
        if dfs:
            return pd.concat(dfs)

    def __get_group_comparison_id(self, groups_str: str, experiment_id: int):
        query = self.session.query(expression_atlas.GroupComparison.id)
        return query.filter_by(group_comparison=groups_str, experiment_id=experiment_id).first().id

    def get_gsea(self, experiment_name: str, groups_str: str, gsea_type: str) -> Optional[pd.DataFrame]:
        """Generate a table of GSEA information for a pair of symbols for a given experiment.

        Parameters
        ----------
        experiment_name : str
            Name of the Expression Atlas experiment.
        groups_str : tuple
            Pairs of gene symbol strings.
        gsea_type : str
            Type of GSEA.

        Returns
        -------
        Returns a pandas DataFrame of GSEA information if file exists matching the passed parameters.
        """
        file_path = os.path.join(
            self.data_dir,
            experiment_name,
            f"{experiment_name}.{groups_str}.{gsea_type}.gsea.tsv",
        )

        if not os.path.exists(file_path):
            return

        df = pd.read_csv(file_path, sep="\t")
        df.columns = [get_standard_name(x) for x in df.columns]

        if "term" in df.columns:
            return df

    def get_log2foldchange(self, experiment_name: str, groups_strs: Tuple[str]) -> pd.DataFrame:
        """Generate a table of log2 fold changes between pairs of gene symbols.

        Parameters
        ----------
        experiment_name : str
            Name of the Expression Atlas experiment.
        groups_strs : tuple
            Pairs of gene symbol strings.

        Returns
        -------
        pandas DataFrame.
        """
        dfs = []
        df_analyticss = self.get_analyticss(experiment_name)
        for df_analytics in df_analyticss:
            for g in groups_strs:
                col_p_value = f"{g}_p_value"
                col_log2foldchange = f"{g}_log2foldchange"
                if {col_p_value, col_log2foldchange}.issubset(df_analytics.columns):
                    cols = ["gene_id", "gene_name", col_p_value, col_log2foldchange]
                    rename_map = {
                        col_p_value: "p_value",
                        col_log2foldchange: "log2foldchange",
                    }

                    if f"{g}_t_statistic" in df_analytics.columns:
                        cols.append(f"{g}_t_statistic")
                        rename_map[f"{g}_t_statistic"] = "t_statistic"

                    df = df_analytics[cols].rename(columns=rename_map)
                    df["group_comparison"] = g
                    dfs.append(df)
        df_concat = pd.concat(dfs)
        return df_concat[
            (df_concat.p_value <= 0.05)
            & df_concat.gene_name.notnull()
            & ((df_concat.log2foldchange <= -1) | (df_concat.log2foldchange >= 1))
        ]

    def get_idf(self, experiment_name: str) -> Optional[pd.DataFrame]:
        """Get Data from IDF by experiment name.

        Parameters
        ----------
        experiment_name : str
            Name of the Expression Atlas experiment.

        Returns
        -------
        DataFrame with IDF values.
        """
        file_path = os.path.join(self.data_dir, experiment_name, f"{experiment_name}.idf.txt")

        if not os.path.exists(file_path):
            return

        rows = []

        for line in open(file_path):
            line_splitted = line.strip().split("\t")

            if len(line_splitted) > 1:
                key_name = get_standard_name(line_splitted[0])
                values = [x.strip() for x in line_splitted[1:] if x.strip()]
                rows.append((key_name, values))

        df = pd.DataFrame(rows, columns=("key_name", "value")).explode("value")
        return df

    def get_sdrf_condensed(self, experiment_name: str) -> Optional[pd.DataFrame]:
        """Generate a condensed version of an experiment's SDRF.

        Parameters
        ----------
        experiment_name : str
            Name of the Expression Atlas experiment.

        Returns
        -------
        pandas DataFrame.
        """
        file_path = os.path.join(self.data_dir, experiment_name, f"{experiment_name}.condensed-sdrf.tsv")

        if not os.path.exists(file_path):
            return

        names = [
            "experiment",
            "method",
            "sample",
            "parameter_type",
            "parameter",
            "value",
            "url",
        ]
        df = pd.read_csv(file_path, sep="\t", header=None, names=names)
        return df

    def get_analyticss(self, experiment_name: str) -> List[pd.DataFrame]:
        """Return list of pandas dataframes detailing specific statistics about a given experiment.

        Parameters
        ----------
        experiment_name : str
            Name of the Expression Atlas experiment.

        Returns
        -------
        List of pandas DataFrames.
        """
        files = os.scandir(os.path.join(self.data_dir, experiment_name))
        analytics_tsv_paths = [f.path for f in files if f.name.endswith("-analytics.tsv")]
        dfs = []
        if analytics_tsv_paths:
            for analytics_tsv_path in analytics_tsv_paths:
                df = pd.read_csv(analytics_tsv_path, sep="\t")
                df.columns = [get_standard_name(x) for x in df.columns]
                dfs.append(df)
        return dfs

    def get_configuration(self, experiment_name: str) -> Optional[DataFrame]:
        """Return pandas dataframe of experiment configuration.

        Parameters
        ----------
        experiment_name : str
            Name of the Expression Atlas experiment.

        Returns
        -------
        pandas DataFrame.
        """
        file_path = os.path.join(self.data_dir, experiment_name, f"{experiment_name}-configuration.xml")

        if not os.path.exists(file_path):
            return

        config = xmltodict.parse(open(file_path).read())["configuration"]

        if config["@experimentType"] != "rnaseq_mrna_baseline":
            compare_dict = {}
            ca = config["analytics"]
            ca_items = ca if isinstance(ca, list) else [ca]

            for item in ca_items:
                for _, contrast in item["contrasts"].items():
                    if isinstance(contrast, (OrderedDict, list)):
                        contrasts = contrast if isinstance(contrast, list) else [contrast]
                        for ind_contrast in contrasts:
                            compare_dict[ind_contrast["@id"]] = ind_contrast["name"]

            return pd.DataFrame(compare_dict.items(), columns=["group_comparison", "name"])



"""This module contains the SQLAlchemy database models that support the definition cache and graph cache."""

import codecs
import datetime
import logging
import os
import re
import urllib
from typing import List, Optional, Tuple, Union
from urllib.parse import quote_plus, urlencode

import pandas as pd
import requests
import sqlalchemy
from lark import Lark, Token, Tree
from sqlalchemy import Boolean, Column, ForeignKey, Index, Integer, String
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.orm import relationship
from sqlalchemy.sql.expression import func
from sqlalchemy_utils import create_database, database_exists
from tqdm import tqdm

from ebel.constants import (FILE, GRAMMAR_NS_ANNO_PATH, GRAMMAR_START_ANNO,
                            GRAMMAR_START_NS, URL)
from ebel.tools import BelRdb

Base = declarative_base()
logger = logging.getLogger(__name__)


def reset_tables(engine: sqlalchemy.engine.Engine, force_new_db: bool) -> None:
    """Drop all tables in database.

    Parameters
    ----------
    engine: sqlalchemy.engine.Engine
        sqlalchemy.engine.Engine
    force_new_db: bool : bool
        True forces to create new tables.
    """
    if not database_exists(engine.url):
        create_database(engine.url)

    if force_new_db:
        Base.metadata.drop_all(bind=engine)
    Base.metadata.create_all(bind=engine, checkfirst=True)


def foreign_key_to(table_name):
    """Create a standard foreign key to a table in the database.

    :param table_name: name of the table without TABLE_PREFIX
    :type table_name: str
    :return: foreign key column
    :rtype: sqlalchemy.Column
    """
    foreign_column = table_name + ".id"
    return Column(Integer, ForeignKey(foreign_column))


class MasterModel(object):
    """Parent class of all models in PyHGNC.

    Automatic creation of table name by class name with project prefix.
    """

    @declared_attr
    def __tablename__(self):
        """Return name of class table."""
        return self.__name__.lower()

    __mapper_args__ = {"always_refresh": True}

    id = Column(Integer, primary_key=True)

    def _to_dict(self):
        """Protected method for converting values to dictionary."""
        data_dict = self.__dict__.copy()
        del data_dict["_sa_instance_state"]
        del data_dict["id"]
        for k, v in data_dict.items():
            if isinstance(v, datetime.date):
                data_dict[k] = data_dict[k].strftime("%Y-%m-%d")
        return data_dict

    def to_dict(self):
        """Return values as dictionary."""
        return self._to_dict()


class Namespace(Base, MasterModel):
    """Represents a BEL Namespace."""

    __tablename__ = "namespace"
    __table_args__ = (Index("idx_url", "url", mysql_length=100),)

    url = Column(String(2048), nullable=False)
    keyword = Column(String(255), index=True)
    cacheable = Column(Boolean)
    case_sensitive = Column(Boolean)

    entries = relationship("NamespaceEntry", back_populates="namespace")


class NamespaceEntry(Base, MasterModel):
    """Represents a name within a BEL namespace."""

    __tablename__ = "namespace_entry"
    __table_args__ = (Index("idx_name", "name", mysql_length=100),)

    name = Column(String(2048), nullable=True)
    encoding = Column(String(8), nullable=True)

    namespace__id = foreign_key_to("namespace")
    namespace = relationship("Namespace", back_populates="entries")


class Annotation(Base, MasterModel):
    """Represents a BEL Annotation."""

    __tablename__ = "annotation"
    __table_args__ = (Index("idx_url2", "url", mysql_length=100),)

    url = Column(String(2048), nullable=False)
    keyword = Column(String(255), index=True)
    cacheable = Column(Boolean)
    case_sensitive = Column(Boolean)

    entries = relationship("AnnotationEntry", back_populates="annotation", cascade="all, delete-orphan")


class AnnotationEntry(Base, MasterModel):
    """Represents a name within a BEL annotation."""

    __tablename__ = "annotation_entry"
    __table_args__ = (Index("idx_identifier", "identifier", mysql_length=100),)

    name = Column(String(2048), nullable=True)
    identifier = Column(String(255), nullable=True)

    annotation__id = foreign_key_to("annotation")
    annotation = relationship("Annotation", back_populates="entries")


class ModelManager:
    """Manager for all models - superclass."""

    def __init__(self, model, entries_model, grammar_start):
        """Init method."""
        self.grammar_start = grammar_start

        self.model = model
        self.entries_model = entries_model
        self.name = model.__tablename__

        self.urls = {}
        self.lists = {}

        self.session = BelRdb().session

        self.errors = []

    def get_entries_not_exists(
        self, keyword: str, url: str, entry_line_column_list
    ) -> List[Tuple[str, int, int, str]]:
        """Get entries in namespace or annotation linked to URL for entry_line_column_list(generator).

        Parameters
        ----------
        keyword : str
            Keyword.
        url : str
            URL string.
        entry_line_column_list : Generator[tuple]
            Generator of tuples with (entry:str , line: int, column: int).

        Returns
        -------
        List[Tuple[str, int, int, str]]
            List of tuples with entry, line, column, hint.

        """
        exists_cache = set()
        not_exists_cache = dict()

        names_not_exists = []

        search_for = self.session.query(self.model.id).filter(self.model.keyword == keyword, self.model.url == url)

        desc = "Check BEL for {}: ".format(keyword)

        for entry, line, column in tqdm(list(entry_line_column_list), desc=desc, ncols=100):
            if entry in not_exists_cache:
                hint = not_exists_cache[entry]
                names_not_exists.append((entry, line, column, hint))

            elif (keyword, url, entry) not in exists_cache:
                exists = search_for.join(self.entries_model).filter(self.entries_model.name == entry).count()

                if exists:
                    exists_cache.update(set([(keyword, url, entry)]))
                else:
                    hint = ""
                    alternatives = (
                        self.session.query(self.entries_model.name, self.model.keyword, self.model.url)
                        .join(self.model)
                        .filter(self.entries_model.name.like(entry))
                        .all()
                    )

                    if alternatives:
                        hint = "Did you mean: "
                        hint += ", ".join([x[1] + ':"' + x[0] + '"(' + x[2] + ")" for x in alternatives])

                    else:
                        if len(entry) >= 6:
                            similars = (
                                self.session.query(self.entries_model.name, self.model.keyword)
                                .join(self.model)
                                .filter(self.entries_model.name.like(entry[:-2] + "%"))
                                .filter(func.length(self.entries_model.name) < len(entry) + 3)
                                .limit(20)
                                .all()
                            )

                            if similars:
                                hint = "Similar: "
                                hint += ", ".join([x[1] + ':"' + x[0] + '"' for x in set(similars)])

                    if not hint:
                        url_query_string = urlencode({"q": entry}, quote_via=quote_plus)
                        hint = "[OLS suggests](https://www.ebi.ac.uk/ols/search?%s)" % url_query_string

                    names_not_exists.append((entry, line, column, hint))
                    not_exists_cache[entry] = hint

        return names_not_exists

    @staticmethod
    def download_url(url: str) -> Tuple[bool, Union[Exception, str]]:
        """Short summary.

        Parameters
        ----------
        url : str
            Description of parameter `url`.

        Returns
        -------
        Tuple[bool, str]
            Description of returned object.

        """
        filename = os.path.basename(urllib.parse.urlparse(url).path)

        try:
            path_to_file = filename
            r = requests.get(url)
            r.raise_for_status()

            open(path_to_file, "wb").write(r.content)

        except requests.exceptions.HTTPError as ex:
            return False, ex

        except requests.exceptions.ConnectionError as ex:
            return False, ex

        except FileNotFoundError as urlex:
            return (
                False,
                f"{str(urlex)}\n{url} does not return a valid belns or belanno file",
            )

        return True, path_to_file

    @staticmethod
    def get_namespace_header(file_path: str) -> Tuple[str, int]:
        """Return header and line number where header ends.

        Parameters
        ----------
        file_path : str
            path to namespace.

        Returns
        -------
        Tuple[str, int]
            header and line number where header ends.

        """
        header = ""
        ends_in_line = 0
        with codecs.open(file_path, "r", encoding="utf-8") as fd:
            for line in fd:
                ends_in_line += 1
                if not re.search(r"^[ \t]*(\[Values\])\s*(\r\n|\r|\n)", line):
                    header += line
                else:
                    break
        return header, ends_in_line

    def save_from_url_or_path(self, keyword: str, url_or_path: str, doc_type: str):
        """Short summary.

        Parameters
        ----------
        keyword : str
            Description of parameter `keyword`.
        url_or_path : str
            Description of parameter `url_or_path`.
        doc_type : str
            Description of parameter `doc_type`.

        Returns
        -------
        type
            Description of returned object.

        """
        path_to_file = None

        if doc_type == URL:
            downloaded, path_to_file_or_error = self.download_url(url_or_path)

            if not downloaded:
                error = path_to_file_or_error
                return False, error

            else:
                path_to_file = path_to_file_or_error

        elif doc_type == FILE:
            path_to_file = url_or_path

        saved, save_error = self.save_in_db(path_to_file=path_to_file, url=url_or_path, keyword=keyword)

        if doc_type == URL:
            os.remove(path_to_file)

        return saved, save_error

    def save_in_db(self, path_to_file: str, url: str, keyword: str) -> Tuple[bool, Optional[Exception]]:
        """Save content of namespace or annotation file from URL with keyword in database.

        Parameters
        ----------
        path_to_file : str
            Description of parameter `path_to_file`.
        url : str
            Description of parameter `url`.
        keyword : str
            Description of parameter `keyword`.

        Returns
        -------
        type
            Description of returned object.
        """
        grammar = load_grammar(GRAMMAR_NS_ANNO_PATH)
        header, header_ends_in_line = self.get_namespace_header(path_to_file)
        lark_parser = Lark(grammar, start=self.grammar_start)

        try:
            tree = lark_parser.parse(header)

        except Exception as e:
            return False, e

        delimiter = first_token_value(tree, "pr_delimiter_string")
        case_sensitive = first_token_value(tree, "pr_case_sensitive_flag")
        cacheable = first_token_value(tree, "pr_cacheable_flag")
        is_case_sensitive = False if re.search("no", case_sensitive, re.I) else False
        is_cacheable = False if re.search("no", cacheable, re.I) else True

        keyword_in_anno = first_token_value(tree, "keyword")
        if keyword != keyword_in_anno:
            warning = f"Keyword {keyword} in BEL namespace URL {url} is different from keyword in BEL script"
            logger.warning(warning)
            # ToDo save in Error classes
            self.errors += [warning]

        model_instance = self.model(
            url=url,
            keyword=keyword,
            cacheable=is_cacheable,
            case_sensitive=is_case_sensitive,
        )
        self.session.add(model_instance)
        self.session.commit()

        table_name = self.name + "_entry"
        second = {"annotation": "identifier", "namespace": "encoding"}
        second_column = second[self.name]

        df = pd.read_csv(
            path_to_file,
            delimiter=delimiter,
            skip_blank_lines=True,
            skiprows=header_ends_in_line,
            names=["name", second_column],
            encoding_errors="ignore",
        )
        df[self.name + "__id"] = model_instance.id
        df.set_index(["name", second_column], inplace=True)

        logger.info(f"Import `{keyword}` table '{table_name}' of engine '{self.session.bind.engine}'")

        df.to_sql(table_name, self.session.bind.engine, if_exists="append", chunksize=1000)
        self.session.commit()
        return True, None

    def keyword_url_exists(self, keyword: str, url: str) -> bool:
        """Check if keyword or annotation exists in the database.

        Parameters
        ----------
        keyword : type
            Description of parameter `keyword`.
        url : type
            Description of parameter `url`.

        Returns
        -------
        type
            Description of returned object.

        """
        result = self.session.query(self.model).filter(self.model.keyword == keyword, self.model.url == url)

        if result.count() == 0:
            exists = False

        else:
            exists = True
            ns_or_anno = result.one()
            if ns_or_anno.cacheable is False:
                self.session.delete(ns_or_anno)
                self.session.commit()
                exists = False

        return exists


class NamespaceManager(ModelManager):
    """Manager for namespaces."""

    def __int__(self, session_obj):
        """Init method."""
        super().__init__(
            model=Namespace,
            entries_model=NamespaceEntry,
            grammar_start=GRAMMAR_START_NS,
        )


class AnnotationManager(ModelManager):
    """Manager for annotations."""

    def __int__(self, session_obj):
        """Init method."""
        super().__init__(
            model=Annotation,
            entries_model=AnnotationEntry,
            grammar_start=GRAMMAR_START_ANNO,
        )


def load_grammar(grammar_path):
    """Return eBNF grammar in lark style.

    Parameters
    ----------
    grammar_path : str
        path to eBNF grammar in lark style.

    Returns
    -------
    string
        eBNF grammar in lark style.

    """
    # FIXME: something to do here
    logger.info("load grammar {}".format(grammar_path))
    with codecs.open(grammar_path, "r", encoding="utf-8") as fd_grammar:
        grammar = fd_grammar.read()

    return grammar


def first_token_value(tree: Tree, subtree_name: str) -> str:
    """Get the first token value of Lark tree with subtree name.

    Parameters
    ----------
    tree : type
        Description of parameter `tree`.
    subtree_name : type
        Description of parameter `subtree_name`.

    Returns
    -------
    type
        Description of returned object.

    """
    # TODO: Get rid of this method by using a Transformer? Is this possible?

    for subtree in tree.iter_subtrees():
        if subtree.data == subtree_name:
            return [node.value for node in subtree.children if isinstance(node, Token)][0]



"""EnsEMBL RDBMS model definition."""

from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()


class Ensembl(Base):
    """Class definition for the ensembl table."""

    __tablename__ = "ensembl"
    id = Column(Integer, primary_key=True)
    enst = Column(String(20), index=True)
    version = Column(Integer)
    chromosome = Column(String(10), index=True)
    start = Column(Integer, index=True)
    stop = Column(Integer, index=True)
    orientation = Column(Integer)
    gene_id = Column(String(255))
    gene_id_short = Column(String(255))
    hgnc_id = Column(String(255), index=True)
    symbol = Column(String(50), index=True)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)



"""GwasCatalog implementation. Depends on HGNC."""

import logging
from typing import Dict

import numpy as np
import pandas as pd
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel.constants import RID
from ebel.manager.orientdb import odb_meta, odb_structure, urls
from ebel.manager.orientdb.biodbs.hgnc import Hgnc
from ebel.manager.orientdb.constants import GWAS_CATALOG
from ebel.manager.rdbms.models import gwas_catalog
from ebel.tools import get_disease_trait_keywords_from_config, get_file_path

logger = logging.getLogger(__name__)


class GwasCatalog(odb_meta.Graph):
    """GWAS Catalog (EBI)."""

    def __init__(
        self,
        client: OrientDB = None,
        disease_trait_keyword: str = "Alzheimer",
        overwrite_config: bool = False,
    ):
        """Init GwasCatalog."""
        self.snp_cache = {}
        self._ensembl_gene_rid_dict: Dict[str:str] = {}
        self.disease_keywords = get_disease_trait_keywords_from_config(
            disease_trait_keyword, overwrite=overwrite_config
        )
        self.client = client
        self.biodb_name = GWAS_CATALOG
        self.url = urls.GWAS_CATALOG
        self.urls = {self.biodb_name: self.url}
        super().__init__(
            nodes=odb_structure.gwascatalog_nodes,
            edges=odb_structure.gwascatalog_edges,
            tables_base=gwas_catalog.Base,
            urls=self.urls,
            biodb_name=self.biodb_name,
        )

    def __len__(self) -> int:
        """Get number of edges in OrientDB."""
        return self.number_of_edges

    def __contains__(self, rs_number: int) -> bool:
        """Checks if RS number (without prefix RS) exists in BEL graph."""
        return self.entry_exists(self.biodb_name, rs_number=rs_number)

    def insert_data(self) -> Dict[str, int]:
        """Insert GwasCatalog data in generic table `gwascatalog`."""
        # TODO: inform data provider about problems in dataset
        df = pd.read_csv(
            get_file_path(self.urls[self.biodb_name], self.biodb_name),
            sep="\t",
            low_memory=False,
            on_bad_lines="warn",
        )
        df.columns = self._standardize_column_names(df.columns)
        df.replace(np.inf, np.nan, inplace=True)  # Get rid of infinity
        # replace non-dates with np.nan
        # non_date = ~df.date_added_to_catalog.str.match(r'^\d{4}-\d{2}-\d{2}$')
        # df.loc[non_date, 'date_added_to_catalog'] = np.nan

        df.index += 1
        df.rename(columns={"snps": "snp"}, inplace=True)
        df.index.rename("id", inplace=True)

        df.upstream_gene_id = df.upstream_gene_id.str.strip()
        df.downstream_gene_id = df.downstream_gene_id.str.strip()

        columns_main_table = [x for x in df.columns if x != "snp_gene_ids"]

        table_name = gwas_catalog.GwasCatalog.__tablename__

        df[columns_main_table].to_sql(table_name, self.engine, if_exists="append")

        df.snp_gene_ids = df.snp_gene_ids.str.strip().str.split(", ")
        df[table_name + "_id"] = df.index

        df_snp_gene_ids = df[[table_name + "_id", "snp_gene_ids"]].explode("snp_gene_ids")
        df_snp_gene_ids.dropna(inplace=True)
        df_snp_gene_ids.index = range(1, df_snp_gene_ids.shape[0] + 1)
        df_snp_gene_ids.rename(columns={"snp_gene_ids": "ensembl_identifier"}, inplace=True)
        df_snp_gene_ids.index.rename("id", inplace=True)
        df_snp_gene_ids.to_sql(gwas_catalog.SnpGene.__tablename__, self.engine, if_exists="append")

        self.session.commit()

        return {self.biodb_name: df.shape[0]}

    def update_bel(self) -> Dict[str, int]:
        """Add human SNPs to OrientDB graph depending on keyword (disease or trait)."""
        hgnc = Hgnc(client=self.client)
        hgnc.update_bel()  # to make sure that all ensembl IDs are available
        logger.info(f"Clear edges for {self.biodb_name}")
        self.clear_edges()
        updated = self.update_interactions()

        if sum(updated.values()) > 0:
            hgnc.update_bel()  # GWAS adds new genes

        return updated

    def update_interactions(self) -> Dict[str, int]:
        """Update mapped, down and upstream has_snp edges."""
        inserted = {}
        logger.info(f"Update interactions for {self.biodb_name}")
        self.clear_edges()
        inserted["has_mapped_snp_gc"] = self.update_mapped_genes()
        inserted["has_downstream_snp_gc"] = self.update_upstream_genes()
        inserted["has_upstream_snp_gc"] = self.update_downstream_genes()
        logger.info(f"Successfully updated interactions for {self.biodb_name}")
        return inserted

    def update_mapped_genes(self) -> int:
        """Update mapped gene information."""
        inserted = 0
        for disease_keyword in self.disease_keywords:
            query_results = (
                self.session.query(gwas_catalog.SnpGene)
                .join(gwas_catalog.GwasCatalog)
                .filter(gwas_catalog.GwasCatalog.disease_trait.like(f"%{disease_keyword}%"))
                .with_entities(
                    gwas_catalog.SnpGene.ensembl_identifier,
                    gwas_catalog.GwasCatalog.snp,
                    gwas_catalog.GwasCatalog.disease_trait,
                    gwas_catalog.GwasCatalog.pubmedid,
                )
                .all()
            )

            inserted += self.insert_snps(query_results, "has_mapped_snp_gc", disease_keyword)
        return inserted

    def update_upstream_genes(self):
        """Update upstream SNP information."""
        inserted = 0

        for disease_keyword in self.disease_keywords:
            query = self.session.query(gwas_catalog.GwasCatalog).filter(
                gwas_catalog.GwasCatalog.disease_trait.like(f"%{disease_keyword}%"),
                gwas_catalog.GwasCatalog.upstream_gene_id.isnot(None),
            )
            query_results = query.with_entities(
                gwas_catalog.GwasCatalog.upstream_gene_id,
                gwas_catalog.GwasCatalog.snp,
                gwas_catalog.GwasCatalog.disease_trait,
                gwas_catalog.GwasCatalog.pubmedid,
            ).all()

            inserted += self.insert_snps(query_results, "has_downstream_snp_gc", disease_keyword)
        return inserted

    def update_downstream_genes(self):
        """Update downstream SNP information."""
        inserted = 0

        for disease_keyword in self.disease_keywords:
            query = self.session.query(gwas_catalog.GwasCatalog).filter(
                gwas_catalog.GwasCatalog.disease_trait.like(f"%{disease_keyword}%"),
                gwas_catalog.GwasCatalog.downstream_gene_id.isnot(None),
            )
            query_results = query.with_entities(
                gwas_catalog.GwasCatalog.downstream_gene_id,
                gwas_catalog.GwasCatalog.snp,
                gwas_catalog.GwasCatalog.disease_trait,
                gwas_catalog.GwasCatalog.pubmedid,
            ).all()

            inserted += self.insert_snps(query_results, "has_upstream_snp_gc", disease_keyword)
        return inserted

    def insert_snps(self, query_results, edge_class: str, keyword: str):
        """Insert SNP metadata into database."""
        columns = ["ensembl_gene_id", "snp", "disease_trait", "pubmedid"]
        snps = pd.DataFrame(query_results, columns=columns)
        snps.set_index("ensembl_gene_id", inplace=True)

        inserted = 0
        desc = f"Update {edge_class} for {keyword}"
        for ensembl_gene_id, row in tqdm(snps.iterrows(), total=snps.shape[0], desc=desc):
            snp_rid = self._get_set_snp_rid(row.snp)
            value_dict = {"disease_trait": row.disease_trait, "pubmed_id": row.pubmedid}

            gene_rid = self._get_gene_rid(ensembl_id=ensembl_gene_id)

            if gene_rid:
                self.create_edge(
                    edge_class,
                    gene_rid,
                    snp_rid,
                    value_dict=value_dict,
                    if_not_exists=True,
                )
                inserted += 1
        return inserted

    def _get_set_snp_rid(self, rs_number: str) -> str:
        """Insert snp (if not exists) and returns OrientDB @rid."""
        results = self.query_class(class_name="snp", limit=1, columns=[], rs_number=rs_number)

        if results:
            rid = results[0][RID]
        else:
            content = {"rs_number": rs_number}
            rid = self.insert_record("snp", content)

        return rid

    @property
    def ensembl_gene_rid_dict(self):
        """Return dict of ensembl IDs as keys and their rIDs as values."""
        if not self._ensembl_gene_rid_dict:
            sql = (
                "Select hgnc.ensembl_gene_id as ensembl, @rid.asString() as rid from gene "
                "where hgnc.ensembl_gene_id IS NOT NULL and namespace='HGNC' and pure=true"
            )
            self._ensembl_gene_rid_dict = {r["ensembl"]: r[RID] for r in [x.oRecordData for x in self.execute(sql)]}
        return self._ensembl_gene_rid_dict

    def _get_gene_rid(self, ensembl_id: str):
        """Insert gene (if not exists) and returns OrientDB @rid."""
        rid = None
        if ensembl_id in self._ensembl_gene_rid_dict:
            rid = self._ensembl_gene_rid_dict[ensembl_id]
        else:
            hgnc_found = self.query_class("hgnc", columns=["symbol"], limit=1, ensembl_gene_id=ensembl_id)
            if hgnc_found:
                r = hgnc_found[0]
                bel = f'g(HGNC:"{r["symbol"]}")'
                data = {
                    "pure": True,
                    "bel": bel,
                    "name": r["symbol"],
                    "namespace": "HGNC",
                    "hgnc": r[RID],
                }
                rid = self.get_create_rid("gene", data, check_for="bel")
                self._ensembl_gene_rid_dict[ensembl_id] = rid
        return rid



"""IntAct API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS


class TestIntAct:
    def test_get_by_uniprot(self, client):
        response = client.get(
            "api/v1/intact?confidence_value=0.56&detection_method=two%20hybrid%20array&detection_method_psimi_id=397&int_a_uniprot_id=P20138&int_b_uniprot_id=Q9Y5Z9&interaction_ids=%25EBI-23994924%25&interaction_type=physical%20association&interaction_type_psimi_id=915&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "confidence_value": 0.56,
            "detection_method": "two hybrid array",
            "detection_method_psimi_id": 397,
            "id": 681219,
            "int_a_uniprot_id": "P20138",
            "int_b_uniprot_id": "Q9Y5Z9",
            "interaction_ids": "intact:EBI-23994924|imex:IM-25472-94969",
            "interaction_type": "physical association",
            "interaction_type_psimi_id": 915,
            "pmid": 32296183,
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results



"""Protein Atlas API methods."""
from ebel.manager.rdbms.models import protein_atlas
from ebel.web.api.ebel.v1 import _get_data


def get_rna_brain_fantom():
    """Get RNA from mouse brain - FANTOM."""
    return _get_data(protein_atlas.ProteinAtlasRnaBrainFantom)


def get_rna_mouse_brain_allen():
    """Get RNA from mouse brain - ALLEN."""
    return _get_data(protein_atlas.ProteinAtlasRnaMouseBrainAllen)


def get_normal_tissue():
    """Get normal tissues."""
    return _get_data(protein_atlas.ProteinAtlasNormalTissue)


def get_subcellular_location():
    """Get subcellular location."""
    return _get_data(protein_atlas.ProteinAtlasSubcellularLocation)


def get_rna_tissue_consensus():
    """Get RNA tissue consensus."""
    return _get_data(protein_atlas.ProteinAtlasRnaTissueConsensus)


def get_rna_brain_gtex():
    """Get RNA Brain GTEX."""
    return _get_data(protein_atlas.ProteinAtlasRnaBrainGtex)



"""API methods."""
import webbrowser

import connexion
from flask_cors import CORS

application = connexion.FlaskApp(__name__)
application.add_api("openapi.yml")

CORS(application.app, expose_headers=["Content-Disposition"])


def run(
    host: str = "0.0.0.0",
    port: int = 5000,
    debug_mode: bool = True,
    open_browser: bool = False,
):
    """Run the API server.

    Parameters
    ----------
    host: str
        Server or host for the API service.
    port: int
        The defined port to run the API server.
    debug_mode: bool
        Whether bebug mode should be enabled.
    open_browser: bool
        If True, automatically opens browser to API UI.
    """
    url = f"http://{host}:{port}/ui"
    if open_browser:
        webbrowser.open(url)
    print(f"Starting web server {url}")
    application.run(host=host, port=port, debug=debug_mode)


if __name__ == "__main__":
    run()



"""eBEL API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS


class TestEbel:
    def test_get_bel_edge_statistics_by_uniprot_accession(self, client):
        # TODO implement when activated in YAML
        pass

    def test_get_intact_by_uniprot(self, client):
        # TODO implement when activated in YAML
        pass

    def test_find_all(self, client):
        response = client.post(
            "api/v1/ebel/find_all",
            json={"term": "CD33"},
        )
        output = format_response_data(response)

        assert isinstance(output, int)
        assert output == 2



"""OrientDB structure.

All vertex, edge and index definitions have to follow the following conversion:

prefix: Used in the hole package for domains like BEL, HGNC, UniProt, ...
typeOfClass or Index
"""
from copy import deepcopy
from enum import Enum
from typing import Dict, List, Optional, Tuple

from ebel.manager.orientdb.odb_defaults import (ODataType, OIndexType,
                                                normalized_pmod)


class OClassType(Enum):
    """Class for constants definitions."""

    NODE = "VERTEX"
    EDGE = "EDGE"
    GENERIC = "GENERIC"


class OProperty(object):
    """Generic class definition for creating properties in the OrientDB database."""

    def __init__(
        self,
        prop_name: str,
        data_type: ODataType,
        linked_class: Optional[str] = None,
        linked_type: Optional[ODataType] = None,
        mandatory: bool = False,
        node_view_label=False,
        node_view_sub_label=False,
    ):
        """Init method."""
        self.prop_name = prop_name
        self.data_type = data_type
        self.linked_type = linked_type
        self.linked_class = linked_class
        self.mandatory = mandatory
        self.node_view_label = node_view_label
        self.node_view_sub_label = node_view_sub_label


class OClass(object):
    """Generic class definition for creating classes in the OrientDB database."""

    def __init__(
        self,
        name: str,
        extends: Tuple[str, ...],
        abstract: bool,
        props: Tuple[OProperty, ...],
        own_class: bool,
        class_type: OClassType,
    ):
        """Init method."""
        self.class_type = class_type
        self.name = name
        self.extends = extends
        self.abstract = abstract
        self.props = list(props)
        self.own_class = own_class

    def add_properties(self, props: List[OProperty]):
        """Add properties to the class."""
        for prop in props:
            self.props.append(prop)
        return self

    def is_edge(self):
        """Check whether class is an edge."""
        return self.class_type == OClassType.EDGE

    def is_node(self):
        """Check whether class is a vertex/node."""
        return self.class_type == OClassType.NODE

    def is_vertex(self):
        """Check whether class is a vertex. Returns "is_node" method and created for convenience."""
        return self.is_node()

    def is_generic(self):
        """Check whether class is generic."""
        return self.class_type == OClassType.GENERIC


class Node(OClass):
    """Generic class definition for creating node classes in the OrientDB database."""

    def __init__(
        self,
        name: str,
        extends: Tuple[OClass, ...] = (),
        abstract: bool = False,
        props: Tuple[OProperty, ...] = (),
        own_class: bool = True,
    ):
        """Init method for Node ODB class."""
        extends = tuple(x.name for x in extends)
        OClass.__init__(self, name, extends, abstract, props, own_class, class_type=OClassType.NODE)


class Edge(OClass):
    """Generic class definition for creating edge classes in the OrientDB database."""

    def __init__(
        self,
        name: str,
        extends: Tuple[OClass, ...] = (),
        abstract: bool = False,
        props: Tuple[OProperty, ...] = (),
        own_class: bool = True,
        in_out: Tuple[Optional[OClass], Optional[OClass]] = (None, None),
    ):
        """Init method for Edge ODB class."""
        extends = tuple(x.name for x in extends)
        in_ = in_out[0].name if in_out[0] else None
        out = in_out[1].name if in_out[1] else None
        OClass.__init__(self, name, extends, abstract, props, own_class, class_type=OClassType.EDGE)
        self.in_out = (in_, out)


class Generic(OClass):
    """Generic class definition for creating generic classes in the OrientDB database."""

    def __init__(
        self,
        name: str,
        extends: Tuple[str, ...] = (),
        abstract: bool = False,
        props: Tuple[OProperty, ...] = (),
        own_class: bool = True,
    ):
        """Init method for Generic ODB class."""
        OClass.__init__(
            self,
            name,
            extends,
            abstract,
            props,
            own_class,
            class_type=OClassType.GENERIC,
        )


class OIndex(object):
    """Generic class definition for creating indices in the OrientDB database."""

    def __init__(self, odb_class: OClass, columns: Tuple[str, ...], index_type: OIndexType):
        """Init method."""
        self.class_name = odb_class.name
        self.columns = columns
        self.index_type = index_type


basic_node = Node(name="V")
basic_edge = Edge(name="E")

##############################################################################
# Definition of BEL vertices, edges and indices
##############################################################################
bel_document = Generic(
    "bel_document",
    props=(
        OProperty("name", ODataType.STRING),
        OProperty("date_uploaded", ODataType.DATETIME),
        OProperty("description", ODataType.STRING),
        OProperty("version", ODataType.STRING),
        OProperty("authors", ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
        OProperty("contact_info", ODataType.STRING),
        OProperty("copyright", ODataType.STRING),
        OProperty("licences", ODataType.STRING),
        OProperty("git", ODataType.EMBEDDEDMAP),
        OProperty("namespaces", ODataType.EMBEDDEDMAP),
        OProperty("annotations", ODataType.EMBEDDEDMAP),
        OProperty("file", ODataType.EMBEDDEDMAP),
        OProperty("git_info", ODataType.EMBEDDEDMAP),
        OProperty("keywords", ODataType.LINKSET, "keyword"),
    ),
)

bel_generics: Tuple[Generic, ...] = (
    Generic(
        "keyword",
        props=(
            OProperty("label", ODataType.STRING),
            OProperty("description", ODataType.STRING),
        ),
    ),
    bel_document,
)
namespace_name = Node(
    "nn",
    (basic_node,),
    abstract=True,
    props=(
        OProperty("namespace", ODataType.STRING),
        OProperty("name", ODataType.STRING),
    ),
)
bel = Node(
    "bel",
    (basic_node,),
    abstract=True,
    props=(
        OProperty("bel", ODataType.STRING, mandatory=True, node_view_sub_label=True),
        OProperty("label", ODataType.STRING),
        OProperty(
            "involved_genes",
            ODataType.EMBEDDEDSET,
            linked_type=ODataType.STRING,
            node_view_label=True,
        ),
        OProperty(
            "involved_other",
            ODataType.EMBEDDEDSET,
            linked_type=ODataType.STRING,
            node_view_label=True,
        ),
        OProperty("suggested_corrections", ODataType.EMBEDDEDMAP, linked_type=ODataType.STRING),
    ),
)
bio_object = Node("bio_object", (bel,), abstract=True, props=(OProperty("chebi", ODataType.INTEGER),))
pure_object = Node(
    "pure_object",
    (basic_node,),
    abstract=True,
    props=(
        OProperty("pure", ODataType.BOOLEAN),
        OProperty("species", ODataType.INTEGER),
    ),
)
genetic_flow = Node("genetic_flow", (bio_object, namespace_name, pure_object), abstract=True)
bio_concept = Node("bio_concept", (bel,), abstract=True)
location_object = Node(
    "location_object",
    (basic_node,),
    abstract=True,
    props=(OProperty("location", ODataType.EMBEDDEDMAP),),
)
bio_act = Node("bio_act", (bel,))
bio_list = Node("bio_list", (bel,), abstract=True)
ebel = Node(
    "ebel",
    (basic_node,),
    abstract=True,
    props=(
        OProperty("bel", ODataType.STRING, node_view_sub_label=True),
        OProperty("name", ODataType.STRING, node_view_label=True),
        OProperty("namespace", ODataType.STRING),
        OProperty("label", ODataType.STRING),
    ),
)
protein = Node(
    "protein",
    (genetic_flow, location_object),
    props=(OProperty("uniprot", ODataType.STRING),),
)
gene = Node("gene", (genetic_flow, location_object))
rna = Node("rna", (genetic_flow, location_object))
abundance = Node("abundance", (bio_object, namespace_name, pure_object))
population = Node("population", (bio_object, namespace_name, pure_object))
from_location = Node("from_location", (ebel,))
to_location = Node("to_location", (ebel,))
biological_process = Node("biological_process", (bio_concept, namespace_name))
pathology = Node("pathology", (bio_concept, namespace_name))
complex_ = Node("complex", (bio_object, namespace_name, pure_object))
micro_rna = Node("micro_rna", (bio_object, namespace_name, location_object))
activity = Node(
    "activity",
    (bio_act, namespace_name),
    props=(OProperty("default", ODataType.STRING),),
)
reaction = Node("reaction", (bio_act,))
degradation = Node("degradation", (bio_act,))
cell_secretion = Node("cell_secretion", (bio_act,))
translocation = Node("translocation", (bio_act,))
cell_surface_expression = Node("cell_surface_expression", (bio_act,))
list_ = Node("list", (bio_list,))
composite = Node("composite", (bio_list,))
variant = Node(
    "variant",
    (ebel,),
    props=(OProperty("hgvs", ODataType.STRING, node_view_label=True),),
)
fragment = Node("fragment", (ebel,))
location = Node("location", (ebel,))
pmod = Node(
    "pmod",
    (ebel,),
    props=(
        OProperty("position", ODataType.INTEGER, node_view_label=True),
        OProperty("type", ODataType.STRING, node_view_label=True),
        OProperty("amino_acid", ODataType.STRING, node_view_label=True),
    ),
)
gmod = Node("gmod", (ebel,))
reactants = Node("reactants", (ebel,))
products = Node("products", (ebel,))
fusion_protein = Node("fusion_protein", (bel,))
fusion_rna = Node("fusion_rna", (bel,))
fusion_gene = Node("fusion_gene", (bel,))

bel_nodes: Tuple[Node, ...] = (
    bel,
    namespace_name,
    pure_object,
    location_object,
    bio_concept,
    biological_process,
    pathology,
    bio_object,
    complex_,
    abundance,
    population,
    genetic_flow,
    gene,
    rna,
    protein,
    micro_rna,
    bio_act,
    activity,
    reaction,
    degradation,
    cell_secretion,
    translocation,
    cell_surface_expression,
    bio_list,
    list_,
    composite,
    ebel,
    variant,
    fragment,
    location,
    pmod,
    gmod,
    from_location,
    to_location,
    reactants,
    products,
    fusion_protein,
    fusion_rna,
    fusion_gene,
)

bel_relation = Edge(
    "bel_relation",
    (basic_edge,),
    abstract=True,
    props=(
        OProperty("evidence", ODataType.STRING),
        OProperty("pmid", ODataType.INTEGER),
        OProperty("pmc", ODataType.STRING),
        OProperty("citation", ODataType.EMBEDDEDMAP),
        OProperty("annotation", ODataType.EMBEDDEDMAP),
        OProperty("document", ODataType.LINKSET, linked_class=bel_document.name),
    ),
)
ebel_relation = Edge("ebel_relation", (basic_edge,), abstract=True)
causal = Edge("causal", (bel_relation,), abstract=True, in_out=(bel, bel))
correlative = Edge("correlative", (bel_relation,), abstract=True, in_out=(bel, bel))
genomic = Edge("genomic", (bel_relation,), abstract=True)
other = Edge("other", (bel_relation,), abstract=True)
deprecated = Edge("deprecated", (bel_relation,), abstract=True)
compiler = Edge("compiler", (bel_relation,), abstract=True)
has_modified = Edge("has_modified", (ebel_relation,), abstract=True)
has_variant_obj = Edge("has_variant_obj", (ebel_relation,), abstract=True)
has_located = Edge("has_located", (ebel_relation,), abstract=True)
has_ppi = Edge(
    "has_ppi",
    (ebel_relation,),
    abstract=True,
    own_class=False,
    in_out=(protein, protein),
)

bel_edges: Tuple[Edge, ...] = (
    bel_relation,
    causal,
    Edge("increases", (causal,)),
    Edge("directly_increases", (causal,)),
    Edge("decreases", (causal,)),
    Edge("directly_decreases", (causal,)),
    Edge("rate_limiting_step_of", (causal,)),
    Edge("causes_no_change", (causal,)),
    Edge("regulates", (causal,)),
    correlative,
    Edge("negative_correlation", (correlative,)),
    Edge("positive_correlation", (correlative,)),
    Edge("association", (correlative,)),
    Edge("no_correlation", (correlative,)),
    genomic,
    Edge("orthologous", (genomic,), in_out=(bel, bel)),
    Edge("transcribed_to", (genomic,), in_out=(rna, gene)),
    Edge("translated_to", (genomic,), in_out=(protein, rna)),
    other,
    Edge("has_member", (other,), in_out=(bel, bel)),
    Edge("has_members", (other,), in_out=(bel, bel)),
    Edge("has_component", (other,), in_out=(bel, bel)),
    Edge("has_components", (other,), in_out=(bel, bel)),
    Edge("equivalent_to", (other,), in_out=(bel, bel)),
    Edge("is_a", (other,), in_out=(bel, bel)),
    Edge("sub_process_of", (other,), in_out=(bel, bel)),
    deprecated,
    Edge("analogous_to", (deprecated,), in_out=(bel, bel)),
    Edge("biomarker_for", (deprecated,), in_out=(bel, bel)),
    Edge("prognostic_biomarker_for", (deprecated,)),
    compiler,
    Edge("acts_in", (compiler,), in_out=(bel, bel)),
    Edge("has_product", (compiler,), in_out=(bel, bel)),
    Edge("has_variant", (compiler,), in_out=(bel, bel)),
    Edge("has_modification", (compiler,), in_out=(bel, bel)),
    Edge("reactant_in", (compiler,), in_out=(bel, bel)),
    Edge("translocates", (compiler,), in_out=(bel, bel)),
    Edge("includes", (compiler,), in_out=(bel, bel)),
    ebel_relation,
    # TODO: check if always basic_node is needed or if we can refine that
    Edge("has__protein", (ebel_relation,), in_out=(protein, basic_node)),
    Edge("has__rna", (ebel_relation,), in_out=(rna, basic_node)),
    Edge("has__gene", (ebel_relation,), in_out=(gene, basic_node)),
    Edge("has__abundance", (ebel_relation,), in_out=(abundance, basic_node)),
    Edge("has__population", (ebel_relation,), in_out=(population, basic_node)),
    Edge("has__location", (ebel_relation,), in_out=(ebel, basic_node)),
    Edge("has__from_location", (ebel_relation,), in_out=(from_location, basic_node)),
    Edge("has__to_location", (ebel_relation,), in_out=(to_location, basic_node)),
    Edge("has__fragment", (ebel_relation,), in_out=(fragment, basic_node)),
    Edge("has__pmod", (ebel_relation,), in_out=(pmod, basic_node)),
    Edge("has__gmod", (ebel_relation,), in_out=(gmod, basic_node)),
    Edge("has__complex", (ebel_relation,), in_out=(complex_, basic_node)),
    Edge("has__micro_rna", (ebel_relation,), in_out=(micro_rna, basic_node)),
    Edge("has__variant", (ebel_relation,), in_out=(variant, basic_node)),
    Edge("has__reactants", (ebel_relation,), in_out=(reactants, basic_node)),
    Edge("has__products", (ebel_relation,), in_out=(products, basic_node)),
    Edge("has__composite", (ebel_relation,), in_out=(composite, basic_node)),
    Edge("has_fragmented_protein", (ebel_relation,), in_out=(bel, bel)),
    has_modified,
    Edge("has_modified_protein", (has_modified,), in_out=(protein, protein)),
    Edge("has_modified_gene", (has_modified,), in_out=(gene, gene)),
    has_variant_obj,
    Edge("has_variant_gene", (has_variant_obj,), in_out=(gene, gene)),
    Edge("has_variant_rna", (has_variant_obj,), in_out=(rna, rna)),
    Edge("has_variant_protein", (has_variant_obj,), in_out=(protein, protein)),
    Edge("has_variant_micro_rna", (has_variant_obj,), in_out=(micro_rna, micro_rna)),
    has_located,
    Edge("has_located_gene", (has_located,), in_out=(gene, gene)),
    Edge("has_located_rna", (has_located,), in_out=(rna, rna)),
    Edge(
        "has_located_protein",
        (has_located,),
        in_out=(protein, protein),
        props=(OProperty("levels", ODataType.EMBEDDEDMAP),),
    ),
    Edge("has_located_micro_rna", (has_located,), in_out=(micro_rna, micro_rna)),
    Edge("has_located_complex", (has_located,), in_out=(complex_, complex_)),
    Edge("has_located_abundance", (has_located,), in_out=(abundance, abundance)),
    Edge("has_located_population", (has_located,), in_out=(population, population)),
    Edge("pathway_interaction", (ebel_relation,), abstract=True),
    has_ppi,
)

bel_indices = (
    OIndex(bel, ("bel",), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(bel, ("involved_genes",), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(bel, ("involved_other",), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(bel_relation, ("evidence",), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(protein, ("uniprot",), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(bel_relation, ("annotation",), OIndexType.DICTIONARY),
    OIndex(bel_relation, ("citation",), OIndexType.DICTIONARY),
    OIndex(bel_relation, ("pmid",), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(namespace_name, ("namespace", "name"), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(namespace_name, ("name",), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(namespace_name, ("namespace",), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(genetic_flow, ("pure",), OIndexType.NOTUNIQUE_HASH_INDEX),
)

##############################################################################
# Definition of IntAct vertices, edges and indices
##############################################################################
has_ppi_ia = Edge(
    "has_ppi_ia",
    (has_ppi,),
    props=(
        OProperty("interaction_type", ODataType.STRING),
        OProperty("interaction_type_psimi_id", ODataType.STRING),
        OProperty("confidence_value", ODataType.FLOAT),
        OProperty("detection_method", ODataType.STRING),
        OProperty("detection_method_psimi_id", ODataType.INTEGER),
        OProperty("pmid", ODataType.INTEGER),
        OProperty("interaction_ids", ODataType.EMBEDDEDMAP),
    ),
)

intact_edges: Tuple[Edge, ...] = (
    has_ppi,
    has_ppi_ia,
)

intact_indices = (OIndex(has_ppi_ia, ("detection_method", "interaction_type"), OIndexType.NOTUNIQUE),)

##############################################################################
# Definition of HGNC generics and indices
# ---------------------------------------
# See .urls.HGNC_JSON
# https://www.genenames.org/help/download
##############################################################################
hgnc = Generic(
    "hgnc",
    props=(
        OProperty("id", ODataType.STRING),
        OProperty("version", ODataType.INTEGER),
        OProperty("alias_name", ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
        OProperty("alias_symbol", ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
        OProperty("bioparadigms_slc", ODataType.STRING),
        OProperty("ccds_id", ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
        OProperty("cd", ODataType.STRING),
        OProperty("cosmic", ODataType.STRING),
        OProperty("date_approved_reserved", ODataType.DATE),
        OProperty("date_modified", ODataType.DATE),
        OProperty("date_name_changed", ODataType.DATE),
        OProperty("date_symbol_changed", ODataType.DATE),
        OProperty("ena", ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
        OProperty("ensembl_gene_id", ODataType.STRING),
        OProperty("entrez_id", ODataType.INTEGER),
        OProperty("enzyme_id", ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
        OProperty("gene_family", ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
        OProperty("gene_family_id", ODataType.EMBEDDEDLIST),
        OProperty("homeodb", ODataType.STRING),
        OProperty("horde_id", ODataType.STRING),
        OProperty("imgt", ODataType.STRING),
        OProperty("iuphar", ODataType.STRING),
        OProperty("kznf_gene_catalog", ODataType.INTEGER),
        OProperty("lncipedia", ODataType.STRING),
        OProperty("lncrnadb", ODataType.STRING),
        OProperty("location", ODataType.STRING),
        OProperty("location_sortable", ODataType.STRING),
        OProperty("locus_group", ODataType.STRING),
        OProperty("locus_type", ODataType.STRING),
        OProperty("lsdb", ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
        OProperty("merops", ODataType.STRING),
        OProperty("mgd_id", ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
        OProperty("mirbase", ODataType.STRING),
        OProperty("name", ODataType.STRING),
        OProperty("omim_id", ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
        OProperty("orphanet", ODataType.INTEGER),
        OProperty("prev_name", ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
        OProperty("prev_symbol", ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
        OProperty("pubmed_id", ODataType.EMBEDDEDLIST, linked_type=ODataType.INTEGER),
        OProperty("refseq_accession", ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
        OProperty("rgd_id", ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
        OProperty("rna_central_ids", ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
        OProperty("snornabase", ODataType.STRING),
        OProperty("status", ODataType.STRING),
        OProperty("symbol", ODataType.STRING),
        OProperty("ucsc_id", ODataType.STRING),
        OProperty("uniprot_ids", ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
        OProperty("uuid", ODataType.STRING),
        OProperty("vega_id", ODataType.STRING),
    ),
)

hgnc_generics: Tuple[Generic, ...] = (hgnc,)

hgnc_nodes: Tuple[Node, ...] = (deepcopy(genetic_flow).add_properties([OProperty("hgnc", ODataType.LINK, "hgnc")]),)

hgnc_indices = (
    OIndex(hgnc, ("symbol",), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(hgnc, ("id",), OIndexType.UNIQUE_HASH_INDEX),
    OIndex(hgnc, ("ensembl_gene_id",), OIndexType.NOTUNIQUE_HASH_INDEX),
)

##############################################################################
# Definition of GwasCatalog generics and indices
##############################################################################

mirtarbase_edges: Tuple[Edge, ...] = (
    Edge(
        "has_mirgene_target",
        (ebel_relation,),
        props=(
            OProperty("support_type", ODataType.STRING),
            OProperty("pmid", ODataType.INTEGER),
            OProperty("experiments", ODataType.EMBEDDEDSET, linked_type=ODataType.STRING),
        ),
        in_out=(rna, micro_rna),
    ),
)

##############################################################################
# Definition of GwasCatalog generics and indices
##############################################################################
snp = Node(
    "snp",
    (ebel,),
    False,
    (OProperty("rs_number", ODataType.STRING, node_view_label=True),),
    own_class=False,
)
gwascatalog_nodes: Tuple[Node, ...] = (snp,)

has_snp = Edge("has_snp", (ebel_relation,), abstract=True, own_class=False, in_out=(snp, gene))
has_mapped_snp = Edge("has_mapped_snp", (has_snp,), abstract=True, own_class=False)
has_downstream_snp = Edge("has_downstream_snp", (has_snp,), abstract=True, own_class=False)
has_upstream_snp = Edge("has_upstream_snp", (has_snp,), abstract=True, own_class=False)
has_snp_gwascatalog = Edge(
    "has_snp_gwascatalog",
    (has_snp,),
    abstract=True,
    props=(
        OProperty("disease_trait", ODataType.STRING),
        OProperty("pubmed_id", ODataType.INTEGER),
    ),
)

gwascatalog_edges: Tuple[Edge, ...] = (
    has_snp,
    has_snp_gwascatalog,
    has_mapped_snp,
    has_downstream_snp,
    has_upstream_snp,
    Edge("has_mapped_snp_gc", (has_mapped_snp, has_snp_gwascatalog)),
    Edge("has_downstream_snp_gc", (has_downstream_snp, has_snp_gwascatalog)),
    Edge("has_upstream_snp_gc", (has_upstream_snp, has_snp_gwascatalog)),
)

##############################################################################
# Definition of Ortholog generics and indices
##############################################################################
drug = Node("drug", (ebel,), abstract=True, own_class=False)
drug_db = Node(
    "drug_db",
    (drug,),
    props=(
        OProperty("label", ODataType.STRING, node_view_label=True),
        OProperty("drugbank_id", ODataType.STRING),
        OProperty("description", ODataType.STRING),
        OProperty("cas_number", ODataType.STRING, node_view_sub_label=True),
        OProperty("indication", ODataType.STRING),
        OProperty("pharmacodynamics", ODataType.STRING),
        OProperty("toxicity", ODataType.STRING),
        OProperty("metabolism", ODataType.STRING),
        OProperty("mechanism_of_action", ODataType.STRING),
    ),
)

drugbank_nodes: Tuple[Node, ...] = (
    drug,
    drug_db,
)

has_drug_target = Edge("has_drug_target", (ebel_relation,), abstract=True, own_class=False)
has_drug_target_db = Edge(
    "has_drug_target_db",
    (has_drug_target,),
    in_out=(protein, drug_db),
    props=(
        OProperty("action", ODataType.STRING),
        OProperty("known_action", ODataType.STRING),
    ),
)

drugbank_edges: Tuple[Edge, ...] = (
    has_drug_target,
    has_drug_target_db,
)

drugbank_indices = (
    OIndex(has_drug_target_db, ("action",), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(drug_db, ("drugbank_id",), OIndexType.UNIQUE_HASH_INDEX),
)

##############################################################################
# Definition of IUPHAR's Guide to Pharmacology drugs
##############################################################################

iuphar_interaction = Edge(
    "iuphar_interaction",
    (ebel_relation,),
    props=(
        OProperty("pmids", ODataType.EMBEDDEDSET, linked_type=ODataType.INTEGER),
        OProperty("assay_description", ODataType.STRING),
        OProperty("affinity_units", ODataType.STRING),
        OProperty("affinity_low", ODataType.FLOAT),
        OProperty("affinity_median", ODataType.FLOAT),
        OProperty("affinity_high", ODataType.FLOAT),
        OProperty("type", ODataType.STRING),
        OProperty("action", ODataType.STRING),
    ),
    in_out=(bel, bel),
)

iuphar_edges: Tuple[Edge, ...] = (
    iuphar_interaction,
    Edge("agonist_of__iu", (iuphar_interaction,)),
    Edge("inhibits__iu", (iuphar_interaction,)),
    Edge("antagonist_of__iu", (iuphar_interaction,)),
    Edge("channel_blocker_of__iu", (iuphar_interaction,)),
    Edge("allosteric_modulator_of__iu", (iuphar_interaction,)),
    Edge("activates__iu", (iuphar_interaction,)),
    Edge("antibody_against__iu", (iuphar_interaction,)),
    Edge("inhibits_gating__iu", (iuphar_interaction,)),
)

##############################################################################
# Definition of Reactome generics and indices
##############################################################################

reactome_nodes: Tuple[Node, ...] = (
    deepcopy(protein).add_properties(
        [OProperty("reactome_pathways", ODataType.EMBEDDEDSET, linked_type=ODataType.STRING)]
    ),
)

##############################################################################
# Definition of Reactome generics and indices
##############################################################################

has_ppi_bg = Edge(
    "has_ppi_bg",
    (has_ppi,),
    abstract=True,
    props=(
        OProperty("modification", ODataType.STRING),
        OProperty("pmids", ODataType.EMBEDDEDSET, linked_type=ODataType.INTEGER),
        OProperty("dois", ODataType.EMBEDDEDSET, linked_type=ODataType.STRING),
        OProperty("biogrid_ids", ODataType.EMBEDDEDSET, linked_type=ODataType.INTEGER),
    ),
)

biogrid_edges: Tuple[Edge, ...] = (
    has_ppi_bg,
    Edge("decreases_bg", (has_ppi_bg,)),
)

biogrid_edges_auto_generated = []
for pmod in normalized_pmod.keys():
    for effect in ["increases", "decreases"]:
        edge_name = f"{effect}_{pmod}_bg"
        biogrid_edges_auto_generated.append(Edge(edge_name, (has_ppi_bg,)))

biogrid_edges += tuple(biogrid_edges_auto_generated)

##############################################################################
# Definition of StringDB generics and indices
##############################################################################

has_action = Edge(
    "has_action",
    (ebel_relation,),
    abstract=True,
    own_class=False,
    in_out=(protein, protein),
)
has_action_st = Edge("has_action_st", (has_action,), abstract=True)
has_ppi_st = Edge(
    "has_ppi_st",
    (has_ppi,),
    abstract=True,
    props=(OProperty("score", ODataType.INTEGER),),
)
controls_expression_of_st = Edge("controls_expression_of_st", (has_action_st,), abstract=True)
increases_expression_of = Edge("increases_expression_of", abstract=True, own_class=False)
decreases_expression_of = Edge("decreases_expression_of", abstract=True, own_class=False)

stringdb_edges: Tuple[Edge, ...] = (
    has_ppi,
    Edge(
        "has_ppi_st",
        (has_ppi,),
        False,
        props=(
            OProperty("neighborhood", ODataType.INTEGER),
            OProperty("neighborhood_transferred", ODataType.INTEGER),
            OProperty("fusion", ODataType.INTEGER),
            OProperty("cooccurence", ODataType.INTEGER),
            OProperty("homology", ODataType.INTEGER),
            OProperty("coexpression", ODataType.INTEGER),
            OProperty("coexpression_transferred", ODataType.INTEGER),
            OProperty("experiments", ODataType.INTEGER),
            OProperty("experiments_transferred", ODataType.INTEGER),
            OProperty("database", ODataType.INTEGER),
            OProperty("database_transferred", ODataType.INTEGER),
            OProperty("textmining", ODataType.INTEGER),
            OProperty("textmining_transferred", ODataType.INTEGER),
            OProperty("combined_score", ODataType.INTEGER),
        ),
    ),
    has_action,
    has_action_st,
    has_ppi_st,
    Edge("activates_st", (has_action_st,)),
    Edge("inhibits_st", (has_action_st,)),
    controls_expression_of_st,
    increases_expression_of,
    decreases_expression_of,
    Edge(
        "increases_expression_of_st",
        (has_action_st, controls_expression_of_st, increases_expression_of),
    ),
    Edge(
        "decreases_expression_of_st",
        (has_action_st, controls_expression_of_st, decreases_expression_of),
    ),
    Edge("controls_pmod_of_st", (has_action_st,)),
)

##############################################################################
# Definition of ClinVar generics and indices
##############################################################################

clinvar_nodes: Tuple[Node, ...] = (snp,)

has_snp_clinvar = Edge(
    "has_snp_clinvar",
    (has_snp,),
    abstract=True,
    props=(
        OProperty("keyword", ODataType.STRING),
        OProperty("clinical_significance", ODataType.STRING),
        OProperty("phenotype", ODataType.STRING),
    ),
)

clinvar_edges: Tuple[Edge, ...] = (
    has_snp,
    has_snp_clinvar,
    has_mapped_snp,
    has_downstream_snp,
    has_upstream_snp,
    Edge("has_mapped_snp_cv", (has_mapped_snp, has_snp_clinvar)),
    Edge("has_downstream_snp_cv", (has_downstream_snp, has_snp_clinvar)),
    Edge("has_upstream_snp_cv", (has_upstream_snp, has_snp_clinvar)),
)

clinvar_indices = (OIndex(snp, ("rs_number",), OIndexType.UNIQUE_HASH_INDEX),)

##############################################################################
# Definition of ClinVar generics and indices
# http://stitch.embl.de
##############################################################################
stitch = Generic(
    "stitch",
    props=(
        OProperty("pubchem_id", ODataType.INTEGER),
        OProperty("ensembl_protein_id", ODataType.INTEGER),
        OProperty("experimental_direct", ODataType.INTEGER),
        OProperty("experimental_transferred", ODataType.INTEGER),
        OProperty("prediction_direct", ODataType.INTEGER),
        OProperty("prediction_transferred", ODataType.INTEGER),
        OProperty("database_direct", ODataType.INTEGER),
        OProperty("database_transferred", ODataType.INTEGER),
        OProperty("textmining_direct", ODataType.INTEGER),
        OProperty("textmining_transferred", ODataType.INTEGER),
        OProperty("combined_score", ODataType.INTEGER),
        OProperty("type", ODataType.STRING),
        OProperty("hgnc", ODataType.LINK, linked_class=hgnc.name),
        OProperty("pubchem", ODataType.LINK, "stitch_pubchem"),
    ),
)

stitch_generics: Tuple[Generic, ...] = (
    Generic(
        "stitch_pubchem",
        props=(
            OProperty("CID", ODataType.INTEGER),
            OProperty("IUPACName", ODataType.STRING),
            OProperty("MolecularFormula", ODataType.STRING),
            OProperty("IsomericSMILES", ODataType.STRING),
            OProperty("InChI", ODataType.STRING),
            OProperty("InChIKey", ODataType.STRING),
        ),
    ),
    stitch,
)

stitch_indices = (
    OIndex(stitch, ("ensembl_protein_id",), OIndexType.NOTUNIQUE),
    OIndex(stitch, ("pubchem_id",), OIndexType.NOTUNIQUE),
    OIndex(stitch, ("pubchem",), OIndexType.NOTUNIQUE),
    OIndex(stitch, ("hgnc",), OIndexType.NOTUNIQUE),
    OIndex(stitch, ("combined_score",), OIndexType.NOTUNIQUE),
)

##############################################################################
# Definition of MirTarBase edges
##############################################################################


##############################################################################
# Definition of DisGeNet
##############################################################################

disgenet_nodes: Tuple[Node, ...] = (snp,)

gene_disease_association = Edge("gene_disease_association", (ebel_relation,), abstract=True)
has_snp_disgenet = Edge(
    "has_snp_disgenet",
    (has_snp,),
    abstract=True,
    props=(
        OProperty("disease_name", ODataType.STRING),
        OProperty("pmid", ODataType.INTEGER),
        OProperty("score", ODataType.FLOAT),
        OProperty("source", ODataType.STRING),
    ),
)

disgenet_edges: Tuple[Edge, ...] = (
    gene_disease_association,
    Edge(
        "disgenet_gene_disease",
        (gene_disease_association,),
        props=(
            OProperty("pmid", ODataType.INTEGER),
            OProperty("score", ODataType.FLOAT),
            OProperty("source", ODataType.STRING),
        ),
    ),
    has_snp,
    has_snp_disgenet,
    has_mapped_snp,
    has_downstream_snp,
    has_upstream_snp,
    Edge("has_mapped_snp_dgn", (has_mapped_snp, has_snp_disgenet)),
    Edge("has_downstream_snp_dgn", (has_downstream_snp, has_snp_disgenet)),
    Edge("has_upstream_snp_dgn", (has_upstream_snp, has_snp_disgenet)),
)

##############################################################################
# Definition of PathwayCommons generics, edges and indices
##############################################################################

pathway_commons_generics: Tuple[Generic, ...] = (
    Generic("pc_pathway_name", props=(OProperty("name", ODataType.STRING),)),
)
has_action_pc = Edge(
    "has_action_pc",
    (has_action,),
    abstract=True,
    props=(
        OProperty("pathways", ODataType.LINKSET, "pc_pathway_name"),
        OProperty("sources", ODataType.EMBEDDEDSET, linked_type=ODataType.STRING),
        OProperty("pmids", ODataType.EMBEDDEDSET, linked_type=ODataType.INTEGER),
        OProperty("type", ODataType.STRING),
    ),
)

pathway_commons_edges: Tuple[Edge, ...] = (
    has_action,
    has_action_pc,
    Edge("controls_expression_of_pc", (has_action_pc,)),
    Edge("controls_phosphorylation_of_pc", (has_action_pc,)),
    Edge("controls_transport_of_pc", (has_action_pc,)),
)

##############################################################################
# Definition of KEGG generics, edges and indices
##############################################################################

pathway_interaction = Edge("pathway_interaction", (ebel_relation,), abstract=True, own_class=False)
has_ppi_kg = Edge(
    "has_ppi_kg",
    (pathway_interaction,),
    abstract=True,
    props=(
        OProperty("interaction_type", ODataType.STRING),
        OProperty("pathway_names", ODataType.EMBEDDEDSET),
    ),
)
increases_expression_of = Edge("increases_expression_of", abstract=True, own_class=False)
decreases_expression_of = Edge("decreases_expression_of", abstract=True, own_class=False)

kegg_edges: Tuple[Edge, ...] = (
    pathway_interaction,
    has_ppi_kg,
    increases_expression_of,
    decreases_expression_of,
    Edge("decreases_pho_kg", (has_ppi_kg,)),
    Edge("increases_gly_kg", (has_ppi_kg,)),
    Edge("increases_me0_kg", (has_ppi_kg,)),
    Edge("increases_pho_kg", (has_ppi_kg,)),
    Edge("increases_ubi_kg", (has_ppi_kg,)),
)

##############################################################################
# NSides definitions
##############################################################################

side_effect = Node(
    "side_effect",
    (ebel,),
    props=(
        OProperty("label", ODataType.STRING, node_view_label=True),
        OProperty("condition_meddra_id", ODataType.STRING, node_view_sub_label=True),
    ),
)

nsides_nodes: Tuple[Node, ...] = (side_effect,)

nsides_edges: Tuple[Edge] = (
    Edge(
        "has_side_effect",
        (ebel_relation,),
        props=(
            OProperty("prr", ODataType.FLOAT),
            OProperty("mean_reporting_frequency", ODataType.FLOAT),
        ),
        in_out=(side_effect, drug_db),
    ),
)

##############################################################################
# All OrientDB classes as Dict
##############################################################################

nodes: Tuple[Node, ...] = (
    bel_nodes
    + hgnc_nodes
    + gwascatalog_nodes
    + drugbank_nodes
    + reactome_nodes
    + clinvar_nodes
    + disgenet_nodes
    + nsides_nodes
)

edges: Tuple[Edge, ...] = (
    bel_edges
    + intact_edges
    + mirtarbase_edges
    + gwascatalog_edges
    + drugbank_edges
    + iuphar_edges
    + biogrid_edges
    + biogrid_edges
    + stringdb_edges
    + clinvar_edges
    + disgenet_edges
    + kegg_edges
    + nsides_edges
)

nodes_and_edges: Tuple[OClass, ...] = nodes + edges

nodes_and_edges_dict: Dict[str, OClass] = {o_class.name: o_class for o_class in nodes_and_edges}
non_abstract_nodes_and_edges_dict: Dict[str, OClass] = {
    o_class.name: o_class for o_class in nodes_and_edges if o_class.abstract is False
}
abstract_nodes_and_edges_dict: Dict[str, OClass] = {
    o_class.name: o_class for o_class in nodes_and_edges if o_class.abstract is True
}


def get_columns(class_name, columns: Tuple[str] = (), exclude_non_serializable: bool = True) -> list:
    """Return list of columns.

    Parameters
    ----------
    class_name : str
        Class name of edge or node.
    columns : List[str], optional
        List of columns.
    exclude_non_serializable : bool
        If True, excludes all non serializable columns.

    Returns
    -------
    list
        List of columns.
    """
    o_class: OClass = nodes_and_edges_dict[class_name]
    exclude_classes = [
        ODataType.LINK,
        ODataType.LINKSET,
        ODataType.LINKMAP,
        ODataType.LINKBAG,
    ]

    columns = list(columns)

    if exclude_non_serializable:
        columns += [p.prop_name for p in o_class.props if p.data_type not in exclude_classes]
    else:
        columns += [p.prop_name for p in o_class.props]
    for extends_name in o_class.extends:
        if extends_name not in ["V", "E"]:
            columns += get_columns(extends_name, tuple(columns), exclude_non_serializable)
    return columns


def get_node_view_labels(name: str, labels: dict = None) -> dict:
    """Get dictionary of labels for a given node.

    Parameters
    ----------
    name : str
        Class name of edge or node.
    labels : dict
        Dictionary of labels and sub labels.

    Returns
    -------
    dict
        {"label": list_of_labels, "sub_label": list_of_sub_labels}
    """
    if not labels:
        labels = {"label": [], "sub_label": []}
    o_class = nodes_and_edges_dict[name]
    for p in o_class.props:
        if p.node_view_label:
            labels["label"].append(p.prop_name)
        if p.node_view_sub_label:
            labels["sub_label"].append(p.prop_name)
    if not labels["label"]:
        for extends_name in o_class.extends:
            if extends_name not in ["V", "E"]:
                get_node_view_labels(extends_name, labels)
    return labels



"""Generic tool definitions for RDBMS submodule."""


def get_or_create(session, model, **kwargs):
    """Create or get (if exists) instance of SQLAlchemy models with kwargs."""
    instance = session.query(model).filter_by(**kwargs).first()
    if instance:
        return instance
    else:
        instance = model(**kwargs)
        return instance



"""DisGeNet."""
import logging
from typing import Dict, Optional

import pandas as pd
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel.manager.orientdb import odb_meta, odb_structure, urls
from ebel.manager.orientdb.constants import DISGENET
from ebel.manager.rdbms.models import disgenet
from ebel.tools import get_disease_trait_keywords_from_config, get_file_path

logger = logging.getLogger(__name__)


class DisGeNet(odb_meta.Graph):
    """DisGeNet (https://www.disgenet.org)."""

    def __init__(self, client: Optional[OrientDB] = None):
        """Init DisGeNet."""
        self.client = client
        self.biodb_name = DISGENET
        self.urls = {
            "disgenet_gene": urls.DISGENET_GDP_ASSOC,
            "disgenet_variant": urls.DISGENET_VDP_ASSOC,
        }
        super().__init__(
            tables_base=disgenet.Base,
            edges=odb_structure.disgenet_edges,
            urls=self.urls,
            biodb_name=self.biodb_name,
        )

        self.disease_keywords = get_disease_trait_keywords_from_config()

    def __len__(self):
        return self.number_of_generics

    def __contains__(self, item):
        # TODO: To be implemented
        return True

    def __repr__(self) -> str:
        """Represent DisGeNet as string."""
        template = "{{BioDatabase:DisGeNet}}[url:{url}, edges:{edges}, generics:{generics}]"
        representation = template.format(url=self.urls, edges=self.number_of_edges, generics=self.number_of_generics)
        return representation

    def insert_data(self) -> Dict[str, int]:
        """Insert data into database."""
        logger.info(f"Import {self.biodb_name.upper()}")
        inserted = dict()
        inserted["sources"] = self._insert_sources()
        inserted["gene_symbols"] = self._insert_gene_symbols()
        inserted["gene_disease_names"] = self._insert_disease_names()
        inserted["gene_disease_pmid_associations"] = self._insert_gene_disease_pmid_associations()
        inserted["variant_disease_pmid_associations"] = self._insert_variant_disease_pmid_associations()
        return inserted

    def __get_file_for_model(self, model):
        """Return filepath of given model."""
        return get_file_path(self.urls[model.__tablename__], self.biodb_name)

    @property
    def file_path_gene(self):
        """Return filepath of gene."""
        return self.__get_file_for_model(disgenet.DisgenetGene)

    @property
    def file_path_variant(self):
        """Return filepath of variant."""
        return self.__get_file_for_model(disgenet.DisgenetVariant)

    def _insert_sources(self):
        df_g = pd.read_csv(self.file_path_gene, sep="\t", usecols=["source"]).drop_duplicates()
        df_v = pd.read_csv(self.file_path_variant, sep="\t", usecols=["source"]).drop_duplicates()
        df = pd.concat([df_g, df_v]).drop_duplicates()
        df.reset_index(inplace=True, drop=True)
        df.index += 1
        df.index.rename("id", inplace=True)
        df.to_sql(disgenet.DisgenetSource.__tablename__, self.engine, if_exists="append")
        return df.shape[0]

    def _insert_disease_names(self) -> int:
        columns_disease = {"diseaseId": "disease_id", "diseaseName": "disease_name"}

        df_gene = (
            pd.read_csv(self.file_path_gene, sep="\t", usecols=list(columns_disease.keys()))
            .rename(columns=columns_disease)
            .drop_duplicates()
            .set_index("disease_id")
        )

        df_variant = (
            pd.read_csv(self.file_path_variant, sep="\t", usecols=list(columns_disease.keys()))
            .rename(columns=columns_disease)
            .drop_duplicates()
            .set_index("disease_id")
        )

        df_concat = pd.concat([df_gene, df_variant]).drop_duplicates()
        df_concat.to_sql(disgenet.DisgenetDisease.__tablename__, self.engine, if_exists="append")
        return df_concat.shape[0]

    def _insert_gene_symbols(self) -> int:
        columns_gene_symols = {"geneId": "gene_id", "geneSymbol": "gene_symbol"}
        df = (
            pd.read_csv(self.file_path_gene, sep="\t", usecols=list(columns_gene_symols.keys()))
            .rename(columns=columns_gene_symols)
            .drop_duplicates()
            .set_index("gene_id")
        )
        df.to_sql(disgenet.DisgenetGeneSymbol.__tablename__, self.engine, if_exists="append")
        return df.shape[0]

    def _merge_with_source(self, df):
        df_sources = pd.read_sql_table(disgenet.DisgenetSource.__tablename__, self.engine).rename(
            columns={"id": "source_id"}
        )
        return pd.merge(df, df_sources, on="source").drop(columns=["source"])

    def _insert_gene_disease_pmid_associations(self) -> int:
        usecols_gene = ["geneId", "diseaseId", "score", "pmid", "source"]
        rename_dict = dict(zip(usecols_gene, self._standardize_column_names(usecols_gene)))
        df = pd.read_csv(self.file_path_gene, sep="\t", usecols=usecols_gene).rename(columns=rename_dict)

        df = self._merge_with_source(df)
        df.index += 1
        df.index.rename("id", inplace=True)
        df.to_sql(disgenet.DisgenetGene.__tablename__, self.engine, if_exists="append")
        return df.shape[0]

    def _insert_variant_disease_pmid_associations(self) -> int:
        usecols_variant = [
            "snpId",
            "chromosome",
            "position",
            "diseaseId",
            "score",
            "pmid",
            "source",
        ]
        rename_dict = dict(zip(usecols_variant, self._standardize_column_names(usecols_variant)))
        df = pd.read_csv(self.file_path_variant, sep="\t", usecols=usecols_variant).rename(columns=rename_dict)

        df = self._merge_with_source(df)
        df.index += 1
        df.index.rename("id", inplace=True)
        df.to_sql(disgenet.DisgenetVariant.__tablename__, self.engine, if_exists="append")

        return df.shape[0]

    def update_interactions(self) -> int:
        """Update edges with disgenet metadata."""
        self.clear_nodes_and_edges()
        # self.update_diseases()
        inserted = self.update_snps()
        self.delete_nodes_with_no_edges("snp")
        return inserted

    def update_diseases(self):
        """Update disease information."""
        # TODO: to be implemented, but disease node class did not exist before and definitions are unclear
        # query = self.session.query(disgenet.DisgenetDisease).filter(
        #     disgenet.DisgenetDisease.disease_name.like('%alzheimer%'))
        # diseases = [x.disease_id for x in query.all()]
        pass

    def update_snps(self) -> int:
        """Update SNP information."""
        snp_type = {
            "mapped": "mapped",
            "downstream": "upstream",
            "upstream": "downstream",
        }
        # TODO: replace SQL with SQL Alchemy statement
        sql_temp = """Select
                snp_id,
                chromosome,
                position,
                disease_name,
                pmid,
                score,
                source
            FROM
                disgenet_variant v INNER JOIN
                disgenet_source s on (v.source_id=s.id) INNER JOIN
                disgenet_disease d on (v.disease_id=d.disease_id)
            WHERE
                disease_name like '%%{}%%' and
                source!='BEFREE'
            GROUP BY
                snp_id,
                chromosome,
                position,
                disease_name,
                pmid,
                score,
                source"""

        results = dict()
        for kwd in self.disease_keywords:
            sql = sql_temp.format(kwd)
            rows = self.engine.execute(sql)
            results[kwd] = rows

        inserted = 0

        snps = {x["rs_number"]: x["rid"] for x in self.query_class("snp", columns=["rs_number"])}

        for trait, kwd_disease_results in results.items():
            for r in tqdm(
                kwd_disease_results,
                desc=f"Update DisGeNET variant interactions for {trait}",
                total=kwd_disease_results.rowcount,
            ):
                snp_id, chromosome, position, disease_name, pmid, score, source = r

                if snp_id in snps:
                    snp_rid = snps[snp_id]

                else:
                    snp_rid = self.insert_record(class_name="snp", value_dict={"rs_number": snp_id})
                    snps[snp_id] = snp_rid

                gene_type_rids = self.get_set_gene_rids_by_position(chromosome, position)

                for gene_type, gene_rids in gene_type_rids.items():
                    for gene_rid in gene_rids:
                        value_dict = {
                            "disease_name": disease_name,
                            "score": score,
                            "source": source,
                            "pmid": pmid,
                        }
                        class_name = f"has_{snp_type[gene_type]}_snp_dgn"
                        self.create_edge(
                            class_name=class_name,
                            from_rid=gene_rid,
                            to_rid=snp_rid,
                            value_dict=value_dict,
                        )
                inserted += 1

        return inserted



# -*- coding: utf-8 -*-

"""This file contains default values for configurations and parameters."""

import logging
import logging.handlers as handlers
import os

from .constants import DATA_DIR, LOG_DIR, PROJECT_DIR

###############################################################################
# UNIPROT taxonomy IDs to import

default_tax_ids = [
    9606,  # Humans
    10116,  # Rats
    10090,  # Mice
    7227,  # Drosophila melanogaster
    694009,
    2697049,  # COVID
    7955,  # Zebrafish
]
###############################################################################

SQLITE_DATABASE_NAME = "ebel.db"
SQLITE_TEST_DATABASE_NAME = "ebel_test.db"
DATABASE_LOCATION = os.path.join(DATA_DIR, SQLITE_DATABASE_NAME)
DEFAULT_TEST_DATABASE_LOCATION = os.path.join(DATA_DIR, SQLITE_TEST_DATABASE_NAME)

###############################################################################
# SQLAlchemy connection strings
# =============================
# SQLite
# ------
CONN_STR_DEFAULT = "sqlite:///" + DATABASE_LOCATION
CONN_STR_TESTS = "sqlite:///" + SQLITE_TEST_DATABASE_NAME
# MySQL
# -----
CONN_STR_MYSQL_PREFIX = "mysql+pymysql://ebel:ebel@localhost/"
CONN_STR_MYSQL = CONN_STR_MYSQL_PREFIX + "ebel?charset=utf8"
CONN_STR_MYSQL_TESTS = CONN_STR_MYSQL_PREFIX + "ebel_test?charset=utf8"

###############################################################################
# Config
config_file_path = os.path.join(PROJECT_DIR, "config.ini")

###############################################################################
# Log Handling
logHandler = handlers.RotatingFileHandler(
    filename=os.path.join(LOG_DIR, "ebel.log"),
    mode="a",
    maxBytes=4098 * 10,  # 4MB file max
    backupCount=0,
)
logh_format = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
logHandler.setFormatter(logh_format)
logHandler.setLevel(logging.DEBUG)

# Console Handler
ch = logging.StreamHandler()
ch_format = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
ch.setFormatter(ch_format)
ch.setLevel(logging.WARNING)



"""ClinicalTrials API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, RESULTS


class TestClinicalTrials:
    def test_get_ct_by_nct_id(self, client):
        response = client.get("api/v1/clinical_trial/by_nct_id?nct_id=NCT00571064", content_type="application/json")

        output = format_response_data(response)
        expected_results = {
            "brief_summary": "This is a study to determine the effectiveness and safety of donepezil hydrochloride (E2020) used to treat residents of assisted living facilities diagnosed with mild, moderate, or severe stage Alzheimer's disease.",
            "brief_title": "The Effectiveness And Safety Of Donepezil Hydrochloride (E2020) In Subjects With Mild To Severe Alzheimer's Disease Residing In An Assisted Living Facility",
            "completion_date": "April 22, 2009",
            "conditions": ["Mild to Severe Alzheimer's Disease"],
            "detailed_description": None,
            "id": 47330,
            "interventions": [{"intervention_name": "Donepezil HCl", "intervention_type": "Drug"}],
            "is_fda_regulated_drug": None,
            "keywords": [],
            "mesh_terms": ["Alzheimer Disease"],
            "nct_id": "NCT00571064",
            "official_title": "A 12-Week, Multicenter, Open Label Study To Evaluate The Effectiveness And Safety Of Donepezil Hydrochloride (E2020) In Subjects With Mild To Severe Alzheimer's Disease Residing In An Assisted Living Facility",
            "org_study_id": "E2020-A001-415",
            "overall_status": "Completed",
            "patient_data_ipd_description": None,
            "patient_data_sharing_ipd": None,
            "phase": "Phase 4",
            "start_date": "January 2008",
            "study_design_intervention_model": "Single Group Assignment",
            "study_design_masking": "None (Open Label)",
            "study_design_primary_purpose": "Treatment",
            "study_type": "Interventional",
        }

        assert isinstance(output, dict)
        assert output == expected_results

    def test_get_ct_by_mesh_term(self, client):
        response = client.get(
            "api/v1/clinical_trial/by_mesh_term?mesh_term=Alzheimer%20Disease&page_size=10&page=1",
            content_type="application/json",
        )

        output = format_response_data(response)

        assert output[NUM_RESULTS] == 1
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit["mesh_term"] == "Alzheimer Disease"
        assert len(hit["nct_ids"]) > 10
        assert "NCT00000171" in hit["nct_ids"]

    def test_get_ct_by_intervention(self, client):
        response = client.get(
            "api/v1/clinical_trial/by_intervention?intervention_name=Donepezil&intervention_type=Drug&page_size=10&page=1",
            content_type="application/json",
        )

        output = format_response_data(response)

        assert output[NUM_RESULTS] == 1
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]

        assert isinstance(hit, dict)
        assert hit["intervention_name"] == "Donepezil"
        assert hit["intervention_type"] == "Drug"
        assert len(hit["nct_ids"]) >= 10
        assert "NCT00000173" in hit["nct_ids"]

    def test_get_ct_by_keyword(self, client):
        response = client.get(
            "api/v1/clinical_trial/by_keyword?keyword=Early%20Alzheimer&page_size=10&page=1",
            content_type="application/json",
        )

        output = format_response_data(response)
        expected_results = {"keyword": "Early Alzheimer", "nct_ids": ["NCT01439555"]}

        assert output[NUM_RESULTS] == 1
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_ct_by_condition(self, client):
        response = client.get(
            "api/v1/clinical_trial/by_condition?condition=Mild%20to%20Severe%20Alzheimer%27s%20Disease&page_size=10&page=1",
            content_type="application/json",
        )

        output = format_response_data(response)
        expected_results = {"condition": "Mild to Severe Alzheimer's Disease", "nct_ids": ["NCT00571064"]}

        assert output[NUM_RESULTS] == 1
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_mesh_term_starts_with(self, client):
        response = client.get(
            "api/v1/clinical_trial/mesh_term/starts_with?mesh_term=alz&page_size=10&page=1",
            content_type="application/json",
        )

        output = format_response_data(response)

        assert output[NUM_RESULTS] == 1
        results = output[RESULTS]
        assert isinstance(results, dict)
        assert "Alzheimer Disease" in results
        assert results["Alzheimer Disease"] > 3000

    def test_get_keyword_starts_with(self, client):
        response = client.get(
            "api/v1/clinical_trial/keyword/starts_with?keyword=alz&page_size=10&page=1",
            content_type="application/json",
        )

        output = format_response_data(response)

        assert output[NUM_RESULTS] >= 9
        results = output[RESULTS]
        assert isinstance(results, dict)
        assert "alzheimer's disease" in results
        assert results["alzheimer's disease"] > 4300

    def test_get_condition_starts_with(self, client):
        response = client.get(
            "api/v1/clinical_trial/condition/starts_with?condition=alz&page_size=10&page=1",
            content_type="application/json",
        )

        output = format_response_data(response)

        assert output[NUM_RESULTS] > 70
        results = output[RESULTS]
        assert isinstance(results, dict)
        assert "Alzheimer Disease 10" in results
        assert results["Alzheimer Disease 10"] > 5990



"""The hub for all things BEL."""
import logging
import os
from collections import namedtuple
from typing import Dict, Iterable, Optional, Set, Union

from pyorientdb.exceptions import PyOrientCommandException
from tqdm import tqdm

from ebel.constants import RID, SPECIES_NAMESPACE
from ebel.manager.orientdb import odb_meta, odb_structure
from ebel.manager.orientdb.biodbs.biogrid import BioGrid
from ebel.manager.orientdb.biodbs.chebi import Chebi
from ebel.manager.orientdb.biodbs.clinical_trials import ClinicalTrials
from ebel.manager.orientdb.biodbs.clinvar import ClinVar
from ebel.manager.orientdb.biodbs.disgenet import DisGeNet
from ebel.manager.orientdb.biodbs.drugbank import DrugBank
from ebel.manager.orientdb.biodbs.ensembl import Ensembl
from ebel.manager.orientdb.biodbs.expression_atlas import ExpressionAtlas
from ebel.manager.orientdb.biodbs.gwas_catalog import GwasCatalog
from ebel.manager.orientdb.biodbs.hgnc import Hgnc
from ebel.manager.orientdb.biodbs.intact import IntAct
from ebel.manager.orientdb.biodbs.iuphar import Iuphar
from ebel.manager.orientdb.biodbs.kegg import Kegg
from ebel.manager.orientdb.biodbs.mirtarbase import MirTarBase
from ebel.manager.orientdb.biodbs.ncbi import Ncbi
from ebel.manager.orientdb.biodbs.nsides import Nsides
from ebel.manager.orientdb.biodbs.pathway_commons import PathwayCommons
from ebel.manager.orientdb.biodbs.protein_atlas import ProteinAtlas
from ebel.manager.orientdb.biodbs.reactome import Reactome
from ebel.manager.orientdb.biodbs.stringdb import StringDb
from ebel.manager.orientdb.biodbs.uniprot import UniProt
from ebel.manager.orientdb.constants import (BIOGRID, CHEBI, CLINICAL_TRIALS,
                                             CLINVAR, DISGENET, DRUGBANK,
                                             ENSEMBL, EXPRESSION_ATLAS,
                                             GWAS_CATALOG, HGNC, INTACT,
                                             IUPHAR, KEGG, MIRTARBASE, NCBI,
                                             NSIDES, PATHWAY_COMMONS,
                                             PROTEIN_ATLAS, REACTOME, STRINGDB,
                                             UNIPROT)
from ebel.manager.orientdb.importer import _BelImporter
from ebel.manager.orientdb.odb_defaults import bel_func_short
from ebel.manager.orientdb.odb_meta import Graph

logger = logging.getLogger(__name__)


class Bel(Graph):
    """Heart and soul of the package, used for interfacing with all of the BioDBs."""

    __hgnc = None
    __uniprot = None
    __drugbank = None
    __gwascatalog = None
    __reactome = None
    __biogrid = None
    __stringdb = None
    __clinical_trials = None
    __intact = None
    __clinvar = None
    __mirtarbase = None
    __expression_atlas = None
    __disgenet = None
    __pathway_commons = None
    __kegg = None
    __ensembl = None
    __iuphar = None
    __chebi = None
    __nsides = None
    __ncbi = None
    __protein_atlas = None

    def __init__(self, graph_config: Optional[dict] = None, overwrite_config: bool = False):
        """Interface for OrientDB database.

        Bel objects allow the user to interact with the different modules in OrientDB and execute queries.

        Parameters
        ----------
        graph_config : Optional[dict]
            If passed, uses the defined parameters for connecting to OrientDB. The dictionary should take the form:
            {
                "user": <ODB username>,
                "password": <ODB password>,
                "db": <ODB database name>,
                "server": <Server ODB is running on (defaults to 'localhost')>,
                "port": <Port ODB is running on (defaults to '2424')
            }
        overwrite_config : bool
            Default False. If True, and graph configuration parameters are passed, overwrites the default values in the
            configuration file with the passed values.
        """
        generics = odb_structure.bel_generics
        nodes = odb_structure.bel_nodes
        edges = odb_structure.bel_edges
        indices = odb_structure.bel_indices
        super().__init__(
            generics,
            nodes,
            edges,
            indices,
            config_params=graph_config,
            overwrite_config=overwrite_config,
        )

    @property
    def protein_atlas(self) -> ProteinAtlas:
        """Create an ProteinAtlas object."""
        if not isinstance(self.__protein_atlas, ProteinAtlas):
            self.__protein_atlas = ProteinAtlas()
        return self.__protein_atlas

    @property
    def ncbi(self) -> Ncbi:
        """Create an Ncbi object."""
        if not isinstance(self.__ncbi, Ncbi):
            self.__ncbi = Ncbi()
        return self.__ncbi

    @property
    def disgenet(self) -> DisGeNet:
        """Create an DisGeNet object."""
        if not isinstance(self.__disgenet, DisGeNet):
            self.__disgenet = DisGeNet()
        return self.__disgenet

    @property
    def expression_atlas(self) -> ExpressionAtlas:
        """Create an ExpressionAtlas object."""
        if not isinstance(self.__expression_atlas, ExpressionAtlas):
            self.__expression_atlas = ExpressionAtlas()
        return self.__expression_atlas

    @property
    def mirtarbase(self) -> MirTarBase:
        """Create an MirTarBase object."""
        if not isinstance(self.__mirtarbase, MirTarBase):
            self.__mirtarbase = MirTarBase()
        return self.__mirtarbase

    @property
    def clinvar(self) -> ClinVar:
        """Create an ClinVar object."""
        if not isinstance(self.__clinvar, ClinVar):
            self.__clinvar = ClinVar()
        return self.__clinvar

    @property
    def hgnc(self) -> Hgnc:
        """Create an Hgnc object."""
        if not isinstance(self.__hgnc, Hgnc):
            self.__hgnc = Hgnc()
        return self.__hgnc

    @property
    def uniprot(self) -> UniProt:
        """Create an UniProt object."""
        if not isinstance(self.__uniprot, UniProt):
            self.__uniprot = UniProt()
        return self.__uniprot

    @property
    def drugbank(self) -> DrugBank:
        """Create an DrugBank object."""
        if not isinstance(self.__drugbank, DrugBank):
            self.__drugbank = DrugBank()
        return self.__drugbank

    @property
    def gwas_catalog(self) -> GwasCatalog:
        """Create an GwasCatalog object."""
        if not isinstance(self.__gwascatalog, GwasCatalog):
            self.__gwascatalog = GwasCatalog()
        return self.__gwascatalog

    @property
    def biogrid(self) -> BioGrid:
        """Create an BioGrid object."""
        if not isinstance(self.__biogrid, BioGrid):
            self.__biogrid = BioGrid()
        return self.__biogrid

    @property
    def stringdb(self) -> StringDb:
        """Create an StringDb object."""
        if not isinstance(self.__stringdb, StringDb):
            self.__stringdb = StringDb()
        return self.__stringdb

    @property
    def reactome(self) -> Reactome:
        """Create an Reactome object."""
        if not isinstance(self.__reactome, Reactome):
            self.__reactome = Reactome()
        return self.__reactome

    @property
    def clinical_trials(self) -> ClinicalTrials:
        """Create an ClinicalTrials object."""
        if not isinstance(self.__clinical_trials, ClinicalTrials):
            self.__clinical_trials = ClinicalTrials()
        return self.__clinical_trials

    @property
    def intact(self) -> IntAct:
        """Create an IntAct object."""
        if not isinstance(self.__intact, IntAct):
            self.__intact = IntAct()
        return self.__intact

    @property
    def pathway_commons(self) -> PathwayCommons:
        """Create an PathwayCommons object."""
        if not isinstance(self.__pathway_commons, PathwayCommons):
            self.__pathway_commons = PathwayCommons()
        return self.__pathway_commons

    @property
    def kegg(self) -> Kegg:
        """Create an Kegg object."""
        if not isinstance(self.__kegg, Kegg):
            self.__kegg = Kegg()
        return self.__kegg

    @property
    def ensembl(self) -> Ensembl:
        """Create an Ensembl object."""
        if not isinstance(self.__ensembl, Ensembl):
            self.__ensembl = Ensembl()
        return self.__ensembl

    @property
    def iuphar(self) -> Iuphar:
        """Create an Iuphar object."""
        if not isinstance(self.__iuphar, Iuphar):
            self.__iuphar = Iuphar()
        return self.__iuphar

    @property
    def chebi(self) -> Chebi:
        """Create an Chebi object."""
        if not isinstance(self.__chebi, Chebi):
            self.__chebi = Chebi()
        return self.__chebi

    @property
    def nsides(self) -> Nsides:
        """Create an NSIDES object."""
        if not isinstance(self.__nsides, Nsides):
            self.__nsides = Nsides()
        return self.__nsides

    def import_json(
        self,
        input_path: Union[str, Iterable[str]],
        extend_graph: bool = True,
        update_from_protein2gene: bool = True,
        skip_drugbank: bool = False,
        drugbank_user: Optional[str] = None,
        drugbank_password: Optional[str] = None,
        include_subfolders: bool = False,
    ) -> list:
        """Import BEL JSON file(s) into OrientDB.

        Parameters
        ----------
        input_path: Iterable or str
            A directory containing BEL JSON files, a single BEL JSON file, or a list of JSON files.
        extend_graph: bool (optional)
            If True, enriches the BEL network after importing. Defaults to True.
        update_from_protein2gene: bool (optional)
            Recursively generates RNA nodes and gene nodes for all protein nodes if none exist. Defaults to True.
        include_subfolders: bool
            Boolean flag to indicate whether to traverse subfolders for BEL files.
        skip_drugbank: bool (optional)
            Flag to disable DrugBank update.
        drugbank_user: str (optional)
            DrugBank user name.
        drugbank_password: str (optional)
            DrugBank password.

        Returns
        -------
        list
            List of files imported
        """
        inserted_files = []
        bel_json_ext = ".bel.json"

        if not skip_drugbank:
            self.drugbank.get_user_passwd(drugbank_user=drugbank_user, drugbank_password=drugbank_password)

        # If directory, get a list of all files there
        if isinstance(input_path, str) and os.path.isdir(input_path):
            if include_subfolders:
                files2d = [
                    [os.path.join(root, f) for f in files if f.endswith(bel_json_ext)]
                    for root, dirs, files in os.walk(input_path)
                ]
                input_path = [x for y in files2d for x in y]
            else:
                input_path = [os.path.join(input_path, file) for file in os.listdir(input_path)]

        if isinstance(input_path, str):
            input_path: Iterable[str] = [input_path]

        for path in input_path:
            if os.path.isfile(path) and path.endswith(bel_json_ext):
                logger.info(f"Begin import: {os.path.basename(path)}")

                try:
                    importer = _BelImporter(path, self.client)
                    file_inserted, number_inserted = importer.insert()
                    if file_inserted:
                        logger.info(f"{os.path.basename(path)} successfully imported")
                        inserted_files.append(path)

                except PyOrientCommandException:
                    logger.error(f"{path} failed to be imported", exc_info=True)

        if inserted_files:
            self._create_and_tag_pure()
            self.update_document_info()

            if update_from_protein2gene:
                self._update_from_protein2gene()

            self._update_involved()

            if extend_graph:
                db_skip = [DRUGBANK] if skip_drugbank else []
                self.enrich_network(skip=db_skip)

            self._update_species()
            self._update_involved()

        return inserted_files

    def enrich_network(
        self,
        include: Union[str, Iterable[str]] = [],
        skip: Union[str, Iterable[str]] = [],
    ) -> Set[str]:
        """Add all external resources to the network.

        Parameters
        ----------
        include: Iterable[str]
            List of databases to include during network enrichment.
        skip: Iterable[str]
            List of databases to skip during enrichment. If a database appears in both "include" and "skip" then
            the database will be skipped.

        Returns
        -------
        Set[str]
            Set of updated databases.
        """
        updated = set()

        biodb_updaters = {
            HGNC: self.hgnc,
            CHEBI: self.chebi,
            ENSEMBL: self.ensembl,
            GWAS_CATALOG: self.gwas_catalog,
            CLINVAR: self.clinvar,
            UNIPROT: self.uniprot,
            REACTOME: self.reactome,
            STRINGDB: self.stringdb,
            INTACT: self.intact,
            BIOGRID: self.biogrid,
            MIRTARBASE: self.mirtarbase,
            PATHWAY_COMMONS: self.pathway_commons,
            DISGENET: self.disgenet,
            KEGG: self.kegg,
            DRUGBANK: self.drugbank,
            IUPHAR: self.iuphar,
            NSIDES: self.nsides,
            CLINICAL_TRIALS: self.clinical_trials,
            PROTEIN_ATLAS: self.protein_atlas,
            NCBI: self.ncbi,
            # EXPRESSION_ATLAS: self.expression_atlas, # TODO: Skipped because file with all results is no longer supported
        }
        # calc all sets (what is used, missing in skip and include)
        include = [] if include == ["[]"] else include  # fixes problems in python3.9/docker
        skip = [] if skip == ["[]"] else skip  # fixes problems in python3.9/docker
        include_set = {include.lower()} if isinstance(include, str) else set([x.lower() for x in include])
        skip_set = {skip.lower()} if isinstance(skip, str) else set([x.lower() for x in skip])
        biodb_set = set([x.lower() for x in biodb_updaters.keys()])
        used_biodb_set = (biodb_set & include_set) if include_set else biodb_set
        used_biodb_set = (used_biodb_set - skip_set) if skip_set else used_biodb_set
        include_not_exists = include_set - biodb_set
        skip_not_exists = skip_set - biodb_set

        logger.info(f"BioDbs for enrichment: {used_biodb_set}")
        if include_not_exists:
            logger.warning(f"{include_not_exists} cannot be included, because they do not exist")
        if skip_not_exists:
            logger.warning(f"{skip_not_exists} cannot be excluded, because they do not exist.")

        # to get the correct order
        db_names = [x for x in biodb_updaters if x in used_biodb_set]

        for db_name in db_names:
            print(f"Enrich network - {db_name.upper()}")
            db: odb_meta.Graph = biodb_updaters[db_name]
            db.update()
            logger.info(f"Enrichment with {db_name} completed.")
            updated.add(db_name)
        self._update_involved()
        return updated

    @property
    def pure_protein_rid_dict(self) -> dict:
        """Get pure protein/rid dictbel. where name is the key and rid is the value."""
        pureprots = self.query_class(class_name="protein", columns=["name"], with_rid=True, pure=True)  # List of dicts
        return {prot["name"]: prot[RID] for prot in pureprots}

    def _update_species(self) -> dict:
        """Add species taxon ID to bel nodes."""
        namespaces_updated = {}

        for molec_state in ["protein", "rna", "gene"]:
            for namespace, species_id in SPECIES_NAMESPACE.items():
                update_pure_bio_sql = f"""UPDATE {molec_state}
                SET species = {species_id}
                WHERE namespace = '{namespace}'"""

                results = self.execute(update_pure_bio_sql)
                namespaces_updated[namespace] = results[0]

        # Update complexes/composites
        results = self.execute("Select @rid.asString() as rid from bel where both('bel_relation').size()>0")
        for r in tqdm(results, desc="Update species"):
            rid = r.oRecordData[RID]
            sql = f"""Select namespace FROM (
                            TRAVERSE out('has__reactants',
                                         'has__products',
                                         'has__protein',
                                         'has__composite',
                                         'has__complex',
                                         'has__gene',
                                         'has__rna') FROM {rid}) WHERE @class in ['protein','rna','gene']"""
            namespace_results = self.execute(sql)
            ns_set = {x.oRecordData["namespace"] for x in namespace_results if namespace_results}

            if len(ns_set) == 1:
                ns = list(ns_set)[0]
                if ns in SPECIES_NAMESPACE.keys():
                    species_sql = f"""UPDATE {rid} SET species = {SPECIES_NAMESPACE[ns]}"""
                    self.execute(species_sql)
                    namespaces_updated[ns] += 1

        return namespaces_updated

    def _update_from_protein2gene(self) -> Dict[str, int]:
        """Adds translated_to and transcribed_to to pure=true proteins and rnas id not exists."""
        added_translated_to = self._add_missing_translated_to_edges()
        added_transcribed_to = self._add_missing_transcribed_to_edges()
        return {
            "added_translated_to": added_translated_to,
            "added_transcribed_to": added_transcribed_to,
        }

    def _create_and_tag_pure(self):
        """Create pure gene, RNA, micro_rna, abundance, complex (as abundance) and protein objects (if not exists).

        Tag all these objects as pure.

        Strategy:
        1. Identify all above mentioned objects with a edges listed below
        2. Check for each object if pure counterpart object exists
        3. If 2=No -> create pure counterpart object
        3. create edge between pure and "modified" objects


        Check for the following modifications (edges):
        out:
        - has__fragment
        - has__variant
        - has__pmod
        - has__location
        in:
        - has_variant
        """
        self._tag_pure()
        self._create_pure_nodes_to_modified()

    def _tag_pure(self) -> int:
        """Tag pure all objects."""
        sql = """Update bio_object set pure = true where
                    @class in ['protein', 'gene', 'rna', 'abundance', 'complex', 'micro_rna'] and
                    outE('has__fragment', 'has__variant', 'has__pmod', 'has__location', 'has__gmod').size() == 0"""
        return len(self.execute(sql))

    def _create_pure_nodes_to_modified(self) -> int:
        """Create all has_modified_(protein|gene) edges in OrientDB (proteins without a pure counterpart)."""
        edge_classes = {
            "has__pmod": "has_modified_protein",
            "has__gmod": "has_modified_gene",
            "has__fragment": "has_fragmented_protein",
            "has__variant": "has_variant_{}",
            "has__location": "has_located_{}",
        }
        results = {}

        sql = """Select
                    out.@rid.asString() as rid,
                    out.@class as class_name,
                    out.name as name,
                    out.namespace as namespace
                from {}"""

        for edge_class in edge_classes:
            results[edge_class] = self.execute(sql.format(edge_class))

        created = 0
        for edge_class, class_name_from_pure in tqdm(edge_classes.items(), desc="Add edges to pure nodes"):
            for row in results[edge_class]:
                r = row.oRecordData

                if r:  # Might return an empty result, but don't know until we look at oRecordData
                    namespace = r["namespace"]
                    name = r["name"]
                    class_name = r["class_name"]

                    if "{}" in class_name_from_pure:
                        cname_from_pure = class_name_from_pure.format(class_name)
                    else:
                        cname_from_pure = class_name_from_pure

                    bel_function = bel_func_short[class_name]
                    bel = f'{bel_function}({namespace}:"{name}")'

                    data = {
                        "namespace": namespace,
                        "name": name,
                        "pure": True,
                        "bel": bel,
                    }

                    from_rid = self.get_create_rid(class_name=class_name, value_dict=data, check_for="bel")
                    to_rid = r[RID]

                    self.create_edge(
                        class_name=cname_from_pure,
                        from_rid=from_rid,
                        to_rid=to_rid,
                        if_not_exists=True,
                    )
                    created += 1

        return created

    def __update_involved(self, node_class) -> Dict[str, int]:
        """Update involved genes and others."""
        sql_rids = f"""Select @rid.asString() as rid from {node_class}
        where involved_genes is null or involved_other is null"""
        results = self.execute(sql_rids)
        updated_involved_genes = 0
        updated_involved_other = 0

        if results:
            for r in tqdm(results, desc=f"Update node constituents for {node_class}"):
                rid = r.oRecordData[RID]
                sql = f"""Update {rid} set involved_genes =
                    (Select name from (
                        traverse out('has__reactants',
                                     'has__products',
                                     'has__protein',
                                     'has__composite',
                                     'has__complex',
                                     'has__gene',
                                     'has__rna') from {rid}) where @class in ['protein','rna','gene'])"""
                updated_involved_genes += self.execute(sql)[0]

                sql = f"""Update {rid} set involved_other =
                    (Select name from (traverse out('has__abundance',
                         'has__reactants',
                         'has__products',
                         'has__composite',
                         'has__complex') from {rid})
                         where @class not in ['protein','gene','rna']
                         and name is not null)"""
                updated_involved_other += self.execute(sql)[0]
        return {
            "updated_involved_genes": updated_involved_genes,
            "updated_involved_other": updated_involved_other,
        }

    def _update_involved(self) -> Dict[str, int]:
        """Update involved genes and others."""
        result = {"updated_involved_genes": 0, "updated_involved_other": 0}
        for node_class in ["bel", "reactants", "products"]:
            sub_result = self.__update_involved(node_class)
            result["updated_involved_genes"] += sub_result["updated_involved_genes"]
            result["updated_involved_other"] += sub_result["updated_involved_other"]
        return result

    def _add_missing_has_variant_edges(self):
        # TODO: Implement this completely
        ModifiedProtein = namedtuple("ModifiedProtein", ["ns", "name", "rids"])
        modified_proteins_sql = """Select
                                    list(@rid.asString()) as rids,
                                    name,
                                    namespace as ns
                                from protein
                                where
                                sum(out('has__fragment').size(),
                                    out('has__variant').size(),
                                    out('has__pmod').size(),
                                    out('has__location').size()
                                ) > 0 and in('has_variant').size() = 0
                                group by name, namespace"""

        results = [r.oRecordData for r in self.execute(modified_proteins_sql)]
        modified_proteins = [ModifiedProtein(r["ns"], r["name"], r["rids"]) for r in results]

        for modified_protein in modified_proteins:
            pass

    def _add_missing_translated_to_edges(self) -> int:
        """Add missing RNAs to proteins and translated_to edges."""
        return self.__add_missing_edges(
            from_class="rna",
            to_class="protein",
            edge_name="translated_to",
            bel_function="r",
        )

    def _add_missing_transcribed_to_edges(self) -> int:
        """Add missing genes to RNAs and transcribed_to edges."""
        return self.__add_missing_edges(
            from_class="gene",
            to_class="rna",
            edge_name="transcribed_to",
            bel_function="g",
        )

    def __add_missing_edges(self, from_class, to_class, edge_name, bel_function) -> int:
        added = 0
        sql_temp = """Select
                    @rid.asString(),
                    namespace,
                    name
                 from
                    {to_class}
                 where
                    pure = true and in('{edge_name}').size() = 0"""

        sql = sql_temp.format(to_class=to_class, edge_name=edge_name)

        results = self.execute(sql)

        if results:
            for to_class_node in tqdm(self.execute(sql), desc=f"Adding {edge_name} edges"):
                p = to_class_node.oRecordData
                bel = '{bel_function}({ns}:"{name}")'.format(
                    ns=p["namespace"], name=p["name"], bel_function=bel_function
                )
                from_rid = self.get_create_rid(
                    from_class,
                    {
                        "namespace": p["namespace"],
                        "name": p["name"],
                        "pure": True,
                        "bel": bel,
                    },
                    check_for="bel",
                )
                self.create_edge(class_name=edge_name, from_rid=from_rid, to_rid=p[RID])
                added += 1

        return added

    def insert_data(self) -> Dict[str, int]:
        """Abstract method."""
        pass

    def update_interactions(self) -> int:
        """Abstract method."""
        pass



"""Pathway Commons RDBMS model definition."""
from sqlalchemy import BigInteger, Column, ForeignKey, Integer, String, Table
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()

pathway_commons__pathway_name = Table(
    "pathway_commons__pathway_name",
    Base.metadata,
    Column("pathway_commons_id", Integer, ForeignKey("pathway_commons.id"), index=True),
    Column(
        "pathway_commons_pathway_name_id",
        Integer,
        ForeignKey("pathway_commons_pathway_name.id"),
        index=True,
    ),
)

pathway_commons__source = Table(
    "pathway_commons__source",
    Base.metadata,
    Column("pathway_commons_id", Integer, ForeignKey("pathway_commons.id"), index=True),
    Column(
        "pathway_commons_source_id",
        Integer,
        ForeignKey("pathway_commons_source.id"),
        index=True,
    ),
)


class PathwayCommons(Base):
    """Class definition for the pathway_commons table."""

    __tablename__ = "pathway_commons"
    id = Column(Integer, primary_key=True)

    participant_a = Column(String(50), index=True)
    interaction_type = Column(String(50), index=True)
    participant_b = Column(String(50), index=True)

    pmids = relationship("Pmid", back_populates="pathway_commons")

    pathway_names = relationship(
        "PathwayName",
        secondary=pathway_commons__pathway_name,
        back_populates="pathway_commonses",
    )

    sources = relationship("Source", secondary=pathway_commons__source, back_populates="pathway_commonses")

    def __str__(self):
        return f"{self.participant_a} {self.interaction_type} {self.participant_b}"

    def as_dict(self):
        """Convert object values to dictionary."""
        pathway_commons = object_as_dict(self)
        pathway_commons.update({"pmids": [x.pmid for x in self.pmids]})
        pathway_commons.update({"pathway_names": [x.name for x in self.pathway_names]})
        pathway_commons.update({"sources": [x.source for x in self.sources]})
        return pathway_commons


class PathwayName(Base):
    """Class definition for the pathway_commons_pathway_name table."""

    __tablename__ = "pathway_commons_pathway_name"
    id = Column(Integer, primary_key=True)

    name = Column(String(255), index=True)

    pathway_commonses = relationship(
        "PathwayCommons",
        secondary=pathway_commons__pathway_name,
        back_populates="pathway_names",
    )

    def __str__(self):
        """Class string definition."""
        return self.name


class Pmid(Base):
    """Class definition for the pathway_commons_pmid table."""

    __tablename__ = "pathway_commons_pmid"
    id = Column(Integer, primary_key=True)

    pmid = Column(BigInteger, index=True)

    pathway_commons_id = Column(Integer, ForeignKey("pathway_commons.id"), index=True)
    pathway_commons = relationship("PathwayCommons", back_populates="pmids")

    def __str__(self):
        """Class string definition."""
        return str(self.pmid)


class Source(Base):
    """Class definition for the pathway_commons_source table."""

    __tablename__ = "pathway_commons_source"
    id = Column(Integer, primary_key=True)

    source = Column(String(50))

    pathway_commonses = relationship("PathwayCommons", secondary=pathway_commons__source, back_populates="sources")

    def __str__(self):
        """Class string definition."""
        return self.source



"""miRTarBase."""
from typing import Dict

import pandas as pd
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel.manager.orientdb import odb_meta, odb_structure, urls
from ebel.manager.orientdb.constants import MIRTARBASE
from ebel.manager.rdbms.models import mirtarbase
from ebel.tools import get_file_path


class MirTarBase(odb_meta.Graph):
    """MirTarBase (from NCBI)."""

    def __init__(self, client: OrientDB = None):
        """Init MirTarBase (http://mirtarbase.mbc.nctu.edu.tw)."""
        self.client = client
        self.biodb_name = MIRTARBASE
        self.urls = {self.biodb_name: urls.MIRTARBASE}
        self.file_path = get_file_path(urls.MIRTARBASE, self.biodb_name)
        super().__init__(
            urls=self.urls,
            tables_base=mirtarbase.Base,
            edges=odb_structure.mirtarbase_edges,
            biodb_name=self.biodb_name,
        )

    def __len__(self) -> Dict[str, int]:
        return self.number_of_generics

    def __contains__(self, item) -> bool:
        # TODO: To be implemented
        return True

    def insert_data(self) -> Dict[str, int]:
        """Insert mirtarbase data into database."""
        # TODO Fix download error -
        #  ssl.SSLError: [SSL: SSLV3_ALERT_HANDSHAKE_FAILURE] sslv3 alert handshake failure (_ssl.c:997)
        df = pd.read_excel(self.file_path)
        df.columns = self._standardize_column_names(df.columns)
        df.index += 1
        df.index.rename("id", inplace=True)

        df.to_sql(
            mirtarbase.Mirtarbase.__tablename__,
            self.engine,
            if_exists="append",
            chunksize=10000,
        )

        return {self.biodb_name: df.shape[0]}

    def update_interactions(self) -> int:
        """Update edges with mirtarbase metadata."""
        self.clear_edges()
        df_symbol_rid = self.get_pure_symbol_rid_df_in_bel_context(class_name="rna", namespace="HGNC")

        sql = f"""Select
                mi_rna,
                target_gene as symbol,
                support_type,
                references_pmid as pmid,
                experiments
            from
                {mirtarbase.Mirtarbase.__tablename__}
            where
                species_mi_rna='Homo sapiens' and
                species_target_gene='Homo sapiens' and
                support_type in ('Functional MTI', 'Non-Functional MTI')"""
        cols = ["mi_rna", "symbol", "support_type", "pmid", "experiments"]
        df_mirtarbase = pd.DataFrame(self.engine.execute(sql).fetchall(), columns=cols)
        df_mirtarbase.experiments = df_mirtarbase.experiments.str.split("//")
        df_join = df_mirtarbase.set_index("symbol").join(df_symbol_rid.set_index("symbol"), how="inner")

        desc = f"Update {self.biodb_name.upper()} interactions"

        updated = 0
        for protein_rid, row in tqdm(df_join.set_index("rid").iterrows(), total=df_join.shape[0], desc=desc):
            mir_data = {
                "bel": f'm(MIRBASE:"{row.mi_rna}")',
                "name": row.mi_rna,
                "namespace": "MIRBASE",
                "pure": True,
            }
            mir_rid = self.get_create_rid("micro_rna", mir_data, check_for="bel")
            value_dict = {
                "support_type": row.support_type,
                "pmid": row.pmid,
                "experiments": row.experiments,
            }
            self.create_edge("has_mirgene_target", mir_rid, str(protein_rid), value_dict=value_dict)
            updated += 1

        return updated



"""KEGG API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS


class TestKegg:
    def test_get_kegg(self, client):
        response = client.get(
            "api/v1/kegg?pathway_identifier=hsa05010&pathway_name=Alzheimer%20disease&kegg_species_id=hsa&kegg_gene_id_a=hsa%3A4137&gene_symbol_a=MAPT&kegg_gene_id_b=hsa%3A100532726&gene_symbol_b=NDUFC2-KCTD14&kegg_int_type=PPrel&interaction_type=inhibition&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "gene_symbol_a": "MAPT",
            "gene_symbol_b": "NDUFC2-KCTD14",
            "id": 95841,
            "interaction_type": "inhibition",
            "kegg_gene_id_a": "hsa:4137",
            "kegg_gene_id_b": "hsa:100532726",
            "kegg_int_type": "PPrel",
            "kegg_species_id": "hsa",
            "pathway_identifier": "hsa05010",
            "pathway_name": "Alzheimer disease",
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_by_gene_symbol(self, client):
        response = client.get(
            "api/v1/kegg/by_gene_symbol?gene_symbol=CD19&page_size=10&page=1", content_type="application/json"
        )
        output = format_response_data(response)
        example_results = {  # Copy/paste example results to get col names quickly
            "gene_symbol_a": "MAPT",
            "gene_symbol_b": "NDUFC2-KCTD14",
            "id": 95841,
            "interaction_type": "inhibition",
            "kegg_gene_id_a": "hsa:4137",
            "kegg_gene_id_b": "hsa:100532726",
            "kegg_int_type": "PPrel",
            "kegg_species_id": "hsa",
            "pathway_identifier": "hsa05010",
            "pathway_name": "Alzheimer disease",
        }
        assert output[NUM_RESULTS] >= 36
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert all([col in hit for col in example_results.keys()])



"""Expression Atlas RDBMS model definition."""
from sqlalchemy import Column, Float, ForeignKey, Integer, String, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()


class Experiment(Base):
    """Table definition for experiment."""

    __tablename__ = "expression_atlas_experiment"

    id = Column(Integer, primary_key=True)

    name = Column(String(100), index=True)
    title = Column(Text)

    idfs = relationship("Idf", back_populates="experiment")
    group_comparisons = relationship("GroupComparison", back_populates="experiment")
    sdrf_condenseds = relationship("SdrfCondensed", back_populates="experiment")

    def as_dict(self):
        """Convert object values to dictionary."""
        experiment = object_as_dict(self)
        experiment.update({"idfs": {idf.key_name: idf.value for idf in self.idfs}})
        gc = [{"groups": x.group_comparison, "name": x.name, "id": x.id} for x in self.group_comparisons]
        experiment.update({"group_comparison": gc})
        return experiment


class Idf(Base):
    """Table definition for IDF."""

    __tablename__ = "expression_atlas_idf"

    id = Column(Integer, primary_key=True)

    key_name = Column(Text, nullable=False)
    value = Column(Text, nullable=False)

    experiment_id = Column(Integer, ForeignKey("expression_atlas_experiment.id"))
    experiment = relationship("Experiment", back_populates="idfs")

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)


class GroupComparison(Base):
    """Table definition for group comparison."""

    __tablename__ = "expression_atlas_group_comparison"

    id = Column(Integer, primary_key=True)

    experiment_id = Column(Integer, ForeignKey("expression_atlas_experiment.id"))
    experiment = relationship("Experiment", back_populates="group_comparisons")

    group_comparison = Column(String(100))
    name = Column(Text)

    fold_changes = relationship("FoldChange", back_populates="group_comparison")
    gseas = relationship("Gsea", back_populates="group_comparison")

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)


class FoldChange(Base):
    """Table definition for fold changes."""

    __tablename__ = "expression_atlas_foldchange"

    id = Column(Integer, primary_key=True)

    gene_id = Column(String(255))
    gene_name = Column(String(100), index=True)
    log2foldchange = Column(Float, index=True)
    p_value = Column(Float, index=True)
    t_statistic = Column(Float)

    group_comparison_id = Column(Integer, ForeignKey("expression_atlas_group_comparison.id"))
    group_comparison = relationship("GroupComparison", back_populates="fold_changes")

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)


class SdrfCondensed(Base):
    """Table definition for SDRF condensed."""

    __tablename__ = "expression_atlas_sdrf_condensed"

    id = Column(Integer, primary_key=True)

    experiment_id = Column(Integer, ForeignKey("expression_atlas_experiment.id"))
    experiment = relationship("Experiment", back_populates="sdrf_condenseds")

    method = Column(String(255))
    sample = Column(String(255))
    parameter_type = Column(String(255))
    parameter = Column(String(255))
    value = Column(String(255))
    url = Column(String(255))

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)


class Gsea(Base):
    """Table definition for Genset enrichment table."""

    __tablename__ = "expression_atlas_gsea"

    id = Column(Integer, primary_key=True)

    group_comparison_id = Column(Integer, ForeignKey("expression_atlas_group_comparison.id"))
    group_comparison = relationship("GroupComparison", back_populates="gseas")

    term = Column(String(255), index=True)
    accession = Column(String(255))
    genes_tot = Column(Integer)
    stat_non_dir_p = Column(Float)
    p_adj_non_dir = Column(Float, index=True)
    significant_in_gene_set = Column(Integer)
    non_significant_in_gene_set = Column(Integer)
    significant_not_in_gene_set = Column(Integer)
    non_significant_not_in_gene_set = Column(Integer)
    effect_size = Column(Float)
    gsea_type = Column(String(100))

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)
